from typing import Optional
from nameguard.models import CheckStatus, Check, GenericCheckResult, NameCheckResult, UNINSPECTED_SKIP_MESSAGE
from nameguard.context import endpoint_name
from nameguard.endpoints import Endpoints
from label_inspector.models import InspectorResult


STATUS = CheckStatus.WARN

# title: Canonical Characters
TITLE_PASS = 'Canonicalization'
TITLE_FAIL = 'Canonicalization'
TITLE_SKIP = 'Canonicalization'

MESSAGE_PASS = 'No signs of impersonation'

MESSAGE_FAIL = 'May be an impersonation of `{}`'
MESSAGE_FAIL_EMOJI = 'Emojis used in this name may be visually confused with other similar emojis'
MESSAGE_FAIL_OTHER = 'May receive potential impersonation warnings'

MESSAGE_SKIP_UNK = 'Contains unknown labels and cannot be checked for impersonation risk'
MESSAGE_SKIP_CANON = 'Contains labels with unknown canonical forms and cannot be checked for impersonation risk'


def check_name(labels: list[Optional[InspectorResult]]) -> GenericCheckResult:
    if None in labels:
        return NameCheckResult(
            check=Check.IMPERSONATION_RISK,
            status=CheckStatus.SKIP,
            _name_message=MESSAGE_SKIP_UNK,
            _title=TITLE_SKIP,
        )
    canonicals = [label.normalized_canonical_label for label in labels]
    if None in canonicals:
        return NameCheckResult(
            check=Check.IMPERSONATION_RISK,
            status=CheckStatus.SKIP,
            _name_message=MESSAGE_SKIP_CANON,
            _title=TITLE_SKIP,
        )
    name = '.'.join(label.label for label in labels)
    canonical = '.'.join(canonicals)
    passed = name == canonical

    if passed:
        message = MESSAGE_PASS
    elif endpoint_name.get() == Endpoints.SECURE_PRIMARY_NAME:
        CHANGED_EMOJI = False
        for label in labels:
            for grapheme in label.graphemes:
                if grapheme.type == 'emoji' and grapheme.confusables_canonical != grapheme.value:
                    CHANGED_EMOJI = True
                    break
        if CHANGED_EMOJI:
            message = MESSAGE_FAIL_EMOJI
        else:
            message = MESSAGE_FAIL.format('.'.join(label.beautiful_canonical_label for label in labels))
    else:
        message = MESSAGE_FAIL_OTHER

    return NameCheckResult(
        check=Check.IMPERSONATION_RISK,
        status=CheckStatus.PASS if passed else STATUS,
        _name_message=message,
        _title=TITLE_PASS if passed else TITLE_FAIL,
    )


UNINSPECTED_SKIP_CHECK_RESULT = NameCheckResult(
    check=Check.IMPERSONATION_RISK,
    status=CheckStatus.SKIP,
    _name_message=UNINSPECTED_SKIP_MESSAGE,
    _title=TITLE_SKIP,
)



from typing import Optional
from pydantic import BaseModel, computed_field, Field, field_serializer
from enum import Enum

from nameguard.generic_utils import capitalize_words
import nameguard.context
from nameguard.endpoints import Endpoints

UNINSPECTED_SKIP_MESSAGE = 'Name is exceptionally long and was not inspected'


class CheckStatus(int, Enum):
    """
    The status of a conducted check.

    * `skip`: This check was skipped because it was not applicable.
    * `info`: This check is informational only.
    * `pass`: This check passed.
    * `warn`: This check failed, this is a minor issue.
    * `alert`: This check failed, this is a major issue.
    """

    INFO = 0
    PASS = 1
    SKIP = 2
    WARN = 3
    ALERT = 4

    @property
    def order(self):
        return self.value

    def __hash__(self) -> int:
        return self.value

    def __str__(self):
        return self.name.lower()


class Rating(int, Enum):
    """
    The rating of a name/label/grapheme based on multiple conducted checks.

    * `pass`: All checks passed.
    * `warn`: At least one check failed with a `WARN` status but no check failed with an `ALERT` status.
    * `alert`: At least one check failed with an `ALERT` status.
    """

    PASS = 2
    WARN = 3
    ALERT = 4

    @property
    def order(self):
        return self.value

    def __hash__(self) -> int:
        return self.value

    def __str__(self):
        return self.name.lower()


class Check(str, Enum):
    """
    The type of a check.

    Common:
    * `normalized`: A name/label/grapheme is ENSIP-15 normalized.

    Grapheme:
    * `confusables`: A grapheme is visually confusable.
    * `invisible`: A grapheme is invisible.
    * `typing_difficulty`: A grapheme is difficult to type.
    * `font_support`: A grapheme is not supported by common fonts.

    Label:
    * `mixed_scripts`: A label contains multiple scripts.
    * `namewrapper_compatible`: TODO
    * `punycode_compatible_label`: A label is compatible with Punycode.
    * `unknown_label`: A label is unknown.

    Name:
    * `impersonation_risk`: A name might be used for impersonation.
    * `punycode_compatible_name`: A name is compatible with Punycode.
    * `namewrapper_fuses`: The NameWrapper configuration of a name is safe.
    * `decentralized_name`: A name is decentralized.
    * `uninspected`: A name is exceptionally long and will not be inspected by NameGuard for performance reasons.
    """

    # Common
    NORMALIZED = 'normalized'

    # Grapheme
    CONFUSABLES = 'confusables'
    INVISIBLE = 'invisible'
    TYPING_DIFFICULTY = 'typing_difficulty'
    FONT_SUPPORT = 'font_support'

    # Label
    MIXED_SCRIPTS = 'mixed_scripts'
    NAMEWRAPPER_COMPATIBLE = 'namewrapper_compatible'
    PUNYCODE_COMPATIBLE_LABEL = 'punycode_compatible_label'
    UNKNOWN_LABEL = 'unknown_label'

    # Name
    IMPERSONATION_RISK = 'impersonation_risk'
    PUNYCODE_COMPATIBLE_NAME = 'punycode_compatible_name'
    NAMEWRAPPER_FUSES = 'namewrapper_fuses'
    DECENTRALIZED_NAME = 'decentralized_name'

    UNINSPECTED = 'uninspected'

    @property
    def human_readable_name(self):
        mapping = {
            self.NORMALIZED: 'Normalized',
            self.CONFUSABLES: 'Recognizable Characters',
            self.INVISIBLE: 'No Hidden Characters',
            self.TYPING_DIFFICULTY: 'Typing Accessibility',
            self.FONT_SUPPORT: 'Font Support',
            self.MIXED_SCRIPTS: 'Single Script',
            self.NAMEWRAPPER_COMPATIBLE: 'NameWrapper Compatible',
            self.PUNYCODE_COMPATIBLE_LABEL: 'DNS Compatible Label',
            self.UNKNOWN_LABEL: 'Known Name',
            self.IMPERSONATION_RISK: 'Canonical Characters',
            self.PUNYCODE_COMPATIBLE_NAME: 'DNS Compatible Name',
            self.NAMEWRAPPER_FUSES: 'NameWrapper Fuses',
            self.DECENTRALIZED_NAME: 'Decentralized Name',
            self.UNINSPECTED: 'Uninspected Name',
        }
        if self in mapping:
            return mapping[self]
        else:
            return capitalize_words(self.value.replace('_', ' '))


def make_severity_dict_from_order(order):
    return {check: len(order) - i for i, check in enumerate(order)}


SEVERITY_DEFAULT = make_severity_dict_from_order(
    [
        # highest severity first
        Check.INVISIBLE,
        Check.NORMALIZED,
        Check.CONFUSABLES,
        Check.TYPING_DIFFICULTY,
        Check.IMPERSONATION_RISK,
        # all other checks get severity 0
    ]
)


SEVERITY_PER_ENDPOINT = {
    Endpoints.SECURE_PRIMARY_NAME: make_severity_dict_from_order(
        [
            Check.IMPERSONATION_RISK,
            Check.INVISIBLE,
            Check.NORMALIZED,
            Check.CONFUSABLES,
            Check.TYPING_DIFFICULTY,
        ]
    ),
}


def get_check_severity(check: Check) -> int:
    endpoint = nameguard.context.endpoint_name.get()
    if endpoint is None:
        return SEVERITY_DEFAULT.get(check, 0)
    else:
        return SEVERITY_PER_ENDPOINT.get(endpoint, SEVERITY_DEFAULT).get(check, 0)


status_rating_mapping = {
    CheckStatus.SKIP: Rating.PASS,
    CheckStatus.INFO: Rating.PASS,
    CheckStatus.PASS: Rating.PASS,
    CheckStatus.WARN: Rating.WARN,
    CheckStatus.ALERT: Rating.ALERT,
}

status_rating = [None] * len(status_rating_mapping)
for status, rating in status_rating_mapping.items():
    status_rating[status] = rating


class GenericCheckResult(BaseModel):
    """
    The result of a conducted check.
    """

    check: Check = Field(examples=[Check.CONFUSABLES])
    status: CheckStatus = Field(examples=[CheckStatus.WARN])

    @field_serializer('status')
    def serialize_status(self, status: CheckStatus, _info):
        return str(status)

    def __init__(self, **data):
        super().__init__(**data)
        self._grapheme_message = data.get('_grapheme_message')
        self._label_message = data.get('_label_message')
        self._name_message = data['_name_message']
        self._title = data['_title']

    _grapheme_message: Optional[str] = None
    _label_message: Optional[str] = None
    _name_message: str
    _title: str

    @property
    def _context(self) -> str:
        raise NotImplementedError

    def raise_context(self) -> 'GenericCheckResult':
        raise NotImplementedError

    @computed_field(description='A message describing the result of the check.')
    @property
    def message(self) -> str:
        ctx = self._context
        if ctx == 'grapheme':
            return self._grapheme_message
        elif ctx == 'label':
            return self._label_message
        else:
            return self._name_message

    @computed_field(description='The human-readable name of the check.')
    @property
    def check_name(self) -> str:
        return self._title

    @property
    def rating(self):
        return status_rating[self.status]

    def __repr__(self):
        return f'{self.check.value}({self.rating})'

    @property
    def order(self):
        """
        Checks are first sorted by status, then by severity.
        Higher risk status always comes first (ALERT > WARN > PASS > INFO > SKIP).
        Within the same status, checks are sorted by severity.
        """
        return (self.status.order, get_check_severity(self.check))

    # Implementing all ops speeds up comparisons

    def __lt__(self, other):
        return self.order < other.order

    def __gt__(self, other):
        return self.order > other.order

    def __eq__(self, other):
        return self.order == other.order

    def __le__(self, other):
        return self.order <= other.order

    def __ge__(self, other):
        return self.order >= other.order

    def __ne__(self, other):
        return self.order != other.order


class GraphemeCheckResult(GenericCheckResult):
    @property
    def _context(self):
        return 'grapheme'

    def raise_context(self):
        return LabelCheckResult(
            check=self.check,
            status=self.status,
            _grapheme_message=self._grapheme_message,
            _label_message=self._label_message,
            _name_message=self._name_message,
            _title=self._title,
        )


class LabelCheckResult(GenericCheckResult):
    @property
    def _context(self):
        return 'label'

    def raise_context(self):
        return NameCheckResult(
            check=self.check,
            status=self.status,
            _grapheme_message=self._grapheme_message,
            _label_message=self._label_message,
            _name_message=self._name_message,
            _title=self._title,
        )


class NameCheckResult(GenericCheckResult):
    @property
    def _context(self):
        return 'name'

    def __hash__(self):
        return hash((self.check, self.status))


UNINSPECTED_CHECK_RESULT = NameCheckResult(
    check=Check.UNINSPECTED,
    status=CheckStatus.ALERT,
    _name_message='Name is exceptionally long and was not inspected',
    _title='Uninspected',
)



from typing import Optional
from label_inspector.models import InspectorResult as Label
from nameguard.models import CheckStatus, Check, LabelCheckResult, NameCheckResult, UNINSPECTED_SKIP_MESSAGE
from label_inspector.common.punycode import puny_analysis, PunycodeCompatibility

STATUS = CheckStatus.WARN

TITLE_PASS = 'DNS Compatibility'
TITLE_FAIL = 'DNS Compatibility'
TITLE_SKIP = 'DNS Compatibility'

L_MESSAGE_PASS = 'Compatible for use with DNS'
N_MESSAGE_PASS = 'Compatible for use with DNS'

L_MESSAGE_FAIL = 'Incompatible for use with DNS'
N_MESSAGE_FAIL = 'Incompatible for use with DNS'

L_MESSAGE_SKIP = 'This label is unknown'
N_MESSAGE_SKIP = 'This name contains unknown labels'


def check_label(label: Optional[Label]) -> LabelCheckResult:
    if label is None:
        return LabelCheckResult(
            check=Check.PUNYCODE_COMPATIBLE_LABEL,
            status=CheckStatus.SKIP,
            _label_message=L_MESSAGE_SKIP,
            _name_message=N_MESSAGE_SKIP,
            _title=TITLE_SKIP,
        )
    passed = label.punycode_compatibility == 'COMPATIBLE'
    return LabelCheckResult(
        check=Check.PUNYCODE_COMPATIBLE_LABEL,
        status=CheckStatus.PASS if passed else STATUS,
        _label_message=L_MESSAGE_PASS if passed else L_MESSAGE_FAIL,
        _name_message=N_MESSAGE_PASS if passed else N_MESSAGE_FAIL,
        _title=TITLE_PASS if passed else TITLE_FAIL,
    )


def check_name(labels: list[Optional[Label]]) -> NameCheckResult:
    if any(label is None for label in labels):
        return NameCheckResult(
            check=Check.PUNYCODE_COMPATIBLE_NAME,
            status=CheckStatus.SKIP,
            _name_message=N_MESSAGE_SKIP,
            _title=TITLE_SKIP,
        )
    labels_passed = all([label.punycode_compatibility == 'COMPATIBLE' for label in labels])
    passed = labels_passed and (
        puny_analysis('.'.join(label.label for label in labels)).compatibility == PunycodeCompatibility.COMPATIBLE
    )
    return NameCheckResult(
        check=Check.PUNYCODE_COMPATIBLE_NAME,
        status=CheckStatus.PASS if passed else STATUS,
        _name_message=N_MESSAGE_PASS if passed else N_MESSAGE_FAIL,
        _title=TITLE_PASS if passed else TITLE_FAIL,
    )


UNINSPECTED_SKIP_CHECK_RESULT = NameCheckResult(
    check=Check.PUNYCODE_COMPATIBLE_NAME,
    status=CheckStatus.SKIP,
    _label_message=UNINSPECTED_SKIP_MESSAGE,
    _name_message=UNINSPECTED_SKIP_MESSAGE,
    _title=TITLE_SKIP,
)



from ens_normalize.normalization import NORMALIZATION
from ens_normalize import is_ens_normalized


def grapheme_is_normalized(g: str) -> bool:
    # optimization for single codepoint graphemes
    if len(g) == 1 and ord(g) in NORMALIZATION.valid:
        return True
    return is_ens_normalized(g)



from typing import Optional
from nameguard.models import CheckStatus, Check, GenericCheckResult, NameCheckResult, UNINSPECTED_SKIP_MESSAGE
from label_inspector.models import InspectorResult

STATUS = CheckStatus.WARN

# title: Decentralized Name
TITLE_PASS = 'Decentralization'
TITLE_FAIL = 'Decentralization'
TITLE_FAIL_UNKNOWN = 'Decentralization'
TITLE_SKIP = 'Decentralization'

MESSAGE_PASS = 'Ownership is decentralized'
MESSAGE_FAIL = 'Ownership is not decentralized'
MESSAGE_FAIL_UNKNOWN = 'Ownership may not be decentralized'

ETH_TLD = 'eth'
DNS_TLD_WHITELIST = {'com', 'net', 'org', 'id', 'io'}


def valid_DNS_tld(label: Optional[str]) -> Optional[bool]:
    """
    Identifies if a label represents a valid TLD in the DNS namespace.
    :param label: the TLD of a name
    :return: True if the label is known to be valid in the DNS namespace. False if the label is
           known to be invalid in the DNS namespace. None if the status of the label is
           unknown in DNS.
    """
    if label == ETH_TLD:
        return False
    elif label == '':
        return True  # DNS root
    elif label in DNS_TLD_WHITELIST:
        return True
    else:
        return None


def namespace_root(labels: list[Optional[str]]) -> str:
    """
    Identifies the namespace root of a name.
    :param labels: The labels of the name to identify the namespace root of.
    :return: `ens` if the namespace root is known to be ENS.
           `dns` if the namespace root is known to be DNS.
           `unknown` if the namespace root is unknown.
    """
    if not labels:
        return 'ens'
    elif len(labels) == 1 and labels[0] == '':
        return 'ens'
    else:
        tld = labels[-1]
        if tld == ETH_TLD:
            return 'ens'
        elif valid_DNS_tld(tld):
            return 'dns'
        else:
            return 'unknown'


def decentralization_status(labels: list[Optional[str]]):
    """
    Identifies the decentralization status of a name.
    :param labels: The labels of the name to identify the decentralization status of.
    :return: `unruggable` if ownership of the name is guaranteed to be decentralized and therefore unruggable.
            `icann` if ownership of the name falls under the jurisdiction of ICANN and therefore is not decentralized.
            `unknown` if the decentralization status of the name is unknown.
    """
    root = namespace_root(labels)
    if root == 'ens':
        if len(labels) > 2:
            return 'unknown'
        else:
            return 'unruggable'
    elif root == 'dns':
        return 'icann'
    elif root == 'unknown':
        return 'unknown'


def check_name(labels: list[Optional[InspectorResult]]) -> GenericCheckResult:
    ds = decentralization_status([label.label if label is not None else None for label in labels])
    if ds == 'unruggable':
        return NameCheckResult(
            check=Check.DECENTRALIZED_NAME,
            status=CheckStatus.PASS,
            _name_message=MESSAGE_PASS,
            _title=TITLE_PASS,
        )
    elif ds == 'icann':
        return NameCheckResult(
            check=Check.DECENTRALIZED_NAME,
            status=CheckStatus.WARN,
            _name_message=MESSAGE_FAIL,
            _title=TITLE_FAIL,
        )
    else:
        return NameCheckResult(
            check=Check.DECENTRALIZED_NAME,
            status=CheckStatus.WARN,
            _name_message=MESSAGE_FAIL_UNKNOWN,
            _title=TITLE_FAIL_UNKNOWN,
        )


UNINSPECTED_SKIP_CHECK_RESULT = NameCheckResult(
    check=Check.DECENTRALIZED_NAME,
    status=CheckStatus.SKIP,
    _name_message=UNINSPECTED_SKIP_MESSAGE,
    _title=TITLE_SKIP,
)



import re
from typing import Optional
from ens.utils import Web3
from hexbytes import HexBytes
from ens.constants import EMPTY_SHA3_BYTES
from urllib.parse import urlparse

from nameguard.exceptions import InvalidNameHash, InvalidTokenID, InvalidEthereumAddress
from nameguard.models import Rating, GenericCheckResult, Check

MAX_INSPECTED_NAME_CHARACTERS = 200  # maximum number of characters to inspect in a name including labels and separators
MAX_INSPECTED_NAME_UNKNOWN_LABELS = 5  # each unknown label in a name is distinctly counted, even if there are multiple instances of the same unknown label in a name.
MAX_NUMBER_OF_NAMES_IN_BULK = 250  # maximum number of names to inspect in a single bulk inspection request


def compute_canonical_from_list(canonicals: list[Optional[str]], sep='') -> Optional[str]:
    """
    Compute a canonical string from a list of canonical strings.
    If any canonical `None`, the result is `None`.
    """
    if any(canonical is None for canonical in canonicals):
        return None
    return sep.join(canonicals)


LABELHASH_REGEX = re.compile(r'^\[[0-9a-f]{64}\]$')
LABELHASH_ETH_REGEX = re.compile(r'^\[[0-9a-f]{64}\]\.eth$')
LABELHASH_IN_NAME_REGEX = re.compile(r'(^|\.)\[[0-9a-f]{64}\]($|\.)')


def is_labelhash_eth(x: str) -> bool:
    return bool(LABELHASH_ETH_REGEX.match(x))


def label_is_labelhash(x: str) -> bool:
    return bool(LABELHASH_REGEX.match(x))


def labelhash_in_name(x: str) -> bool:
    return bool(LABELHASH_IN_NAME_REGEX.match(x))


def hexbytes_to_int(hb: HexBytes) -> int:
    return int(hb.hex(), base=16)


def int_to_hexstr(n: int, hex_len=64) -> str:
    """
    Given an integer `n`, return a hex-string prefixed with '0x',
    padded with 0s to have exactly `hex_len` digits.

    Parameters
    ----------
    n : int
    hex_len: int, optional
    Returns
    -------
    str
        `n` in hex-string format (padded with 0s to match 64 digits and prefixed with 0x).
    Raises
    ------
    ValueError
        If `n` in hex repr has more digits than `hex_len`.
    """
    res = f'{n:#0{hex_len + 2}x}'
    if len(res) > hex_len + 2:
        raise ValueError(f'Resulting hex number has more digits ({len(res) - 2}) than specified ({hex_len}).')
    return res


def labelhash_from_label(label: str) -> str:
    return Web3().keccak(text=label).hex()


def namehash_from_name(name: str) -> str:
    """
    Calculate the namehash of a name.
    Detects labelhashes in the name.
    """
    node = EMPTY_SHA3_BYTES
    if name:
        labels = name.split('.')
        for label in reversed(labels):
            if label_is_labelhash(label):
                labelhash = HexBytes(label[1:-1])
            else:
                labelhash = Web3().keccak(text=label)
            node = Web3().keccak(node + labelhash)
    return node.hex()


def namehash_from_labelhash(labelhash_hexstr: str, parent_name='eth') -> str:
    parent_namehash_hexstr = namehash_from_name(parent_name)
    node = Web3().keccak(HexBytes(parent_namehash_hexstr) + HexBytes(labelhash_hexstr))
    return node.hex()


def validate_token_id(token_id: str) -> str:
    """
    Validate token_id string and return normalized token_id.

    Parameters
    ----------
    token_id : str
        A string representing a token_id. It can be in
            a) decimal format - decimal integer of any length,
            b) hex format - hex digits prefixed with 0x.
    Returns
    -------
    str
        normalized input
    Raises
    ------
    InvalidTokenID
    """
    token_id = token_id.lower()
    if token_id.startswith('0x'):
        if not validate_hex(token_id, require_prefix=True, required_length=None):
            raise InvalidTokenID('Invalid hex number.')
        return token_id
    elif token_id.isdigit():
        return token_id
    else:
        raise InvalidTokenID('Must be a valid, decimal integer.')


def validate_namehash(namehash: str) -> str:
    """
    Validate namehash string and return namehash in hex-string format.

    Parameters
    ----------
    namehash : str
        A string representing a namehash. It can be in
            a) decimal format - decimal integer of any length,
            b) hex format - 64 hex digits prefixed with 0x.
    Returns
    -------
    str
        Namehash in hex-string format (padded with 0s to match 64 digits and prefixed with 0x).
    Raises
    ------
    InvalidNameHash
    """
    namehash = namehash.lower()
    if namehash.startswith('0x'):
        if not validate_hex(namehash, require_prefix=True, required_length=66):
            raise InvalidNameHash("Hex number must be 64 digits long and prefixed with '0x'.")
        return namehash
    else:
        try:
            int_namehash = int(namehash)
        except ValueError:
            raise InvalidNameHash(
                "Must be a valid, decimal integer or a hex number with 64 digits, prefixed with '0x'."
            )
        try:
            hex_namehash = int_to_hexstr(int_namehash)
        except ValueError:
            raise InvalidNameHash('The decimal integer converted to base-16 should have at most 64 digits.')
        return hex_namehash


def validate_ethereum_address(address: str) -> str:
    """
    Validate address string and return normalized address.

    Parameters
    ----------
    address : str
        A string representing a address. It must be in hex format - hex digits prefixed with 0x.
    Returns
    -------
    str
        normalized input
    Raises
    ------
    InvalidEthereumAddress
    """
    address = address.lower()
    if not validate_hex(address, require_prefix=True, required_length=42):
        raise InvalidEthereumAddress("Hex number must be 40 digits long and prefixed with '0x'.")
    return address


def validate_hex(hex: str, *, require_prefix: bool = True, required_length: int = None) -> bool:
    """
    Hex string must be lowercased.
    """
    if required_length is not None and len(hex) != required_length:
        return False
    if require_prefix:
        if not hex.startswith('0x'):
            return False
        return all(c in '0123456789abcdef' for c in hex[2:])
    else:
        return all(c in '0123456789abcdef' for c in hex)


def calculate_nameguard_rating(check_results: list[GenericCheckResult]) -> Rating:
    return max((check.rating for check in check_results), default=Rating.PASS)


def count_risks(check_results: list[GenericCheckResult]) -> int:
    return sum(1 for check in check_results if check.rating > Rating.PASS)


def agg_checks(check_results: list[GenericCheckResult]) -> list[GenericCheckResult]:
    out: dict[Check, GenericCheckResult] = {}
    for result in check_results:
        out[result.check] = max(out.get(result.check, result), result)
    return list(out.values())


def get_highest_risk(check_results: list[GenericCheckResult]) -> Optional[GenericCheckResult]:
    return max((check for check in check_results if check.rating > Rating.PASS), default=None)


def detect_grapheme_link_name(link: str) -> str:
    """
    Possible link formats:

    * https://unicodeplus.com/U+{...}
    * https://unicode.link/inspect/utf8:{...}
    * http://📙.la/{...}
    """
    host = urlparse(link).hostname
    if host == 'unicodeplus.com':
        return 'UnicodePlus'
    elif host == 'unicode.link':
        # title of th epage
        return 'Unicode Visualizer'
    elif host == '📙.la':
        return 'Emojipedia'
    else:
        return 'Unknown External Page'



from typing import Optional
from label_inspector.models import InspectorResult
from nameguard.models import CheckStatus, Check, GenericCheckResult, LabelCheckResult, UNINSPECTED_SKIP_MESSAGE

STATUS = CheckStatus.ALERT

# title: Known Name
TITLE_PASS = 'Label Inspectability'
TITLE_FAIL = 'Label Inspectability'
TITLE_SKIP = 'Label Inspectability'

L_MESSAGE_PASS = 'This label is known'
N_MESSAGE_PASS = 'This name is known'

L_MESSAGE_FAIL = 'This label is unknown'
N_MESSAGE_FAIL = 'This name contains unknown labels'


def check_label(label: Optional[InspectorResult]) -> GenericCheckResult:
    passed = label is not None
    return LabelCheckResult(
        check=Check.UNKNOWN_LABEL,
        status=CheckStatus.PASS if passed else STATUS,
        _label_message=L_MESSAGE_PASS if passed else L_MESSAGE_FAIL,
        _name_message=N_MESSAGE_PASS if passed else N_MESSAGE_FAIL,
        _title=TITLE_PASS if passed else TITLE_FAIL,
    )


UNINSPECTED_SKIP_CHECK_RESULT = LabelCheckResult(
    check=Check.UNKNOWN_LABEL,
    status=CheckStatus.SKIP,
    _label_message=UNINSPECTED_SKIP_MESSAGE,
    _name_message=UNINSPECTED_SKIP_MESSAGE,
    _title=TITLE_SKIP,
)



import logging


logger = logging.getLogger('nameguard')
logger.setLevel(logging.DEBUG)
stream_handler = logging.StreamHandler()
stream_handler.setFormatter(logging.Formatter('%(asctime)s - %(levelname)s - %(message)s'))
stream_handler.setLevel(logging.DEBUG)
logger.addHandler(stream_handler)



from typing import Optional, Literal
from pydantic import BaseModel, Field, computed_field, field_serializer
from enum import Enum
from ens_normalize import ens_beautify

from nameguard.context import endpoint_name
from nameguard.models.checks import GenericCheckResult, Rating, Check, UNINSPECTED_CHECK_RESULT

from nameguard.utils import detect_grapheme_link_name
from nameguard.endpoints import Endpoints


class Normalization(str, Enum):
    """
    The ENSIP-15 normalization status of a name or label.

    * `normalized`: The name or label is normalized.
    * `unnormalized`: The name or label is not normalized.
    * `unknown`: The name or label is unknown because it cannot be looked up from its hash.
    """

    NORMALIZED = 'normalized'
    UNNORMALIZED = 'unnormalized'
    UNKNOWN = 'unknown'


class GraphemeNormalization(str, Enum):
    """
    The ENSIP-15 normalization status of a grapheme.
    This check does not consider the context of the grapheme and is **not** equivalent to `ens_normalize(grapheme)`.
    A normalized grapheme can be combined with other normalized graphemes to form an unnormalized label.
    The position of a grapheme in a label can also affect the normalization status of the label.

    * `normalized`: The grapheme is normalized.
    * `unnormalized`: The grapheme is not normalized.
    """

    NORMALIZED = 'normalized'
    UNNORMALIZED = 'unnormalized'


class ConsolidatedReport(BaseModel):
    """
    The consolidated inspection results of all the risks and limitations found within a name, label, or grapheme.
    """

    rating: Rating = Field(examples=[Rating.WARN])

    @field_serializer('rating')
    def serialize_rating(self, rating: Rating, _info):
        return str(rating)

    @computed_field(description='A human-readable title based on the `rating`.')
    @property
    def title(self) -> str:
        if self.rating is Rating.PASS:
            return 'Looks Good'
        elif self.rating is Rating.WARN:
            if (
                self.highest_risk is not None
                and self.highest_risk.check is Check.IMPERSONATION_RISK
                and endpoint_name.get() == Endpoints.SECURE_PRIMARY_NAME
            ):
                return 'Impersonation Risk'
            else:
                return 'Some Risk'
        else:  # self.rating is Rating.ALERT:
            if endpoint_name.get() == 'fake-ens-name-check':
                return 'Fake ENS Name'
            else:
                return 'High Risk'

    @computed_field(description='A human-readable subtitle based on the `rating`.')
    @property
    def subtitle(self) -> str:
        if self.rating is Rating.PASS:
            return 'All security checks passed!'
        elif self.rating is Rating.WARN:
            return 'Review risks before proceeding'
        else:  # self.rating is Rating.ALERT:
            return f'Better not to use this {self._string_type}'

    risk_count: int = Field(
        description='The number of checks that have a status of `alert` or `warn`.',
        examples=[2],
    )

    highest_risk: Optional[GenericCheckResult] = Field(
        description='The check considered to be the highest risk. If no check has a status of `alert` or `warn`, this field is `null`.',
    )

    @property
    def _string_type(self) -> str:
        raise NotImplementedError

    @property
    def _string_value(self) -> str:
        raise NotImplementedError


class ConsolidatedGraphemeGuardReport(ConsolidatedReport):
    """
    Grapheme analysis result.
    """

    normalization: GraphemeNormalization

    grapheme: str = Field(
        description='The analyzed grapheme.',
        examples=['v'],
    )

    grapheme_name: str = Field(
        description='The name of the grapheme.',
        examples=['Latin Small Letter V'],
    )

    grapheme_type: str = Field(
        description='The type of the grapheme. If all characters in the grapheme have the same type, that type is returned. Otherwise, `special` is returned.\n'
        '* `simple_letter` - `[a-z]`\n'
        '* `simple_number` - `[0-9]`\n'
        '* `other_letter` - a letter (single-char grapheme) in any script that is not simple; `LC` class http://www.unicode.org/reports/tr44/#GC_Values_Table \n'
        '* `other_number` - a digit (single-char grapheme) in any script that is not simple; `N` class http://www.unicode.org/reports/tr44/#GC_Values_Table \n'
        '* `hyphen` - a hyphen\n'
        '* `dollarsign` - a dollar sign\n'
        '* `underscore` - an underscore\n'
        '* `emoji` - an emoji or emoji ZWJ sequence\n'
        '* `invisible` - zero width joiner or non-joiner\n'
        "* `special` - for any grapheme that doesn't match one of the other classifications or if characters have different types",
        examples=['simple_letter'],
    )

    grapheme_script: str = Field(
        description='Script name of the grapheme computed from the script names of its characters.',
        examples=['Latin'],
    )

    grapheme_description: str = Field(
        description='Description of the grapheme type.',
        examples=['A-Z letter'],
    )

    unicode_version: Optional[str] = Field(
        description='The Unicode version in which the grapheme was introduced or assigned.\n'
        '* `null` if the grapheme is not assigned to any specific Unicode version',
        examples=['15.0'],
    )

    @property
    def _string_type(self) -> str:
        return 'grapheme'

    @property
    def _string_value(self) -> str:
        return self.grapheme


class LabelGuardReport(ConsolidatedReport):
    """
    Label analysis result.
    """

    label: str = Field(
        description='The analyzed label. If the label is unknown, this field is `[labelhash]`.',
        examples=['vitalìk', '[99b91f5ec34a22cf0fb21c9f43be6c6417d9991e979c1dca532a8e74d1feec23]'],
    )

    labelhash: str = Field(
        description='The labelhash of the label in hex format prefixed with `0x`.',
        examples=['0x41459971a5b847d0eef67bc11e2845f1074a2aed1ac6a357378b18fba50816e4'],
    )

    normalization: Normalization

    @computed_field(description='Beautified version of `label`.')
    @property
    def beautiful_label(self) -> Optional[str]:
        return ens_beautify(self.label) if self.normalization is Normalization.NORMALIZED else None

    checks: list[GenericCheckResult] = Field(
        description='A list of checks that were performed on the label.',
    )

    graphemes: Optional[list[ConsolidatedGraphemeGuardReport]] = Field(
        description='A list of graphemes that were analyzed in the label. If the label is unknown, this field is `null`.',
    )

    canonical_label: Optional[str] = Field(
        description='The canonical form of the analyzed label.\n'
        '* `null` if the canonical form of any grapheme is not known\n'
        '* `null` if the result would be unnormalized, even if the canonical form of all graphemes is known\n'
        '* `[labelhash]` if the label is unknown',
        examples=['vitalik'],
    )

    @property
    def _string_type(self) -> str:
        return 'label'

    @property
    def _string_value(self) -> str:
        return self.label


class ConsolidatedNameGuardReport(ConsolidatedReport):
    """
    The consolidated inspection results of all the risks and limitations NameGuard found within a name.

    Generated through a detailed inspection of all labels and graphemes within `name`.

    A `ConsolidatedNameGuardReport` consolidates the results of all `checks` on all labels and graphemes in a
    `NameGuardReport` into a `ConsolidatedReport` without the need to explicitly return all
    the details of the `NameGuardReport`.
    """

    name: str = Field(
        description='The name that NameGuard inspected. Some labels in this name may be represented as "[labelhash]" '
        'if and only if all of the following is true:\n'
        '\n'
        '1. The query sent to NameGuard when requesting the report represented the label as a "[labelhash]".\n'
        '2. The decoded label of "[labelhash]" was not found within the ENS Subgraph for the specified `network`.',
        examples=['vitalìk.eth', '[af498306bb191650e8614d574b3687c104bc1cd7e07c522954326752c6882770].eth'],
    )

    namehash: str = Field(
        description='The ENSIP-1 namehash of the name in hex format prefixed with 0x.',
        examples=['0xd48fd5598e605861cbd8e45419b41b83739bff52eaef0e283181bbe0a43a5b32'],
    )

    normalization: Normalization

    inspected: bool = Field(
        description='Whether or not the name was inspected. If `false` then the name was exceptionally long and was not inspected for performance reasons',
        examples=[True],
    )

    @computed_field(description='Beautified version of `name`.')
    @property
    def beautiful_name(self) -> Optional[str]:
        return ens_beautify(self.name) if self.normalization is Normalization.NORMALIZED else None

    @property
    def _string_type(self) -> str:
        return 'name'

    @property
    def _string_value(self) -> str:
        return self.name


class NameGuardReport(ConsolidatedNameGuardReport):
    """
    NameGuard report that contains the full results of all `checks` on all `labels` in a name.
    """

    checks: list[GenericCheckResult] = Field(
        description='The results of all checks performed by NameGuard on `name`.',
    )

    labels: Optional[list[LabelGuardReport]] = Field(
        description='The analyzed labels of the name.\n' '* `null` if and only if `inspected` field is false',
    )

    canonical_name: Optional[str] = Field(
        description='The canonical form of the analyzed name.\n'
        '* `null` if the canonical form of any label is not known or `inspected` field is false\n'
        '* `can contain labelhashes when some labels are unknown`',
        examples=['vitalik.eth'],
    )


class UninspectedNameGuardReport(NameGuardReport):
    """
    Uninspected name analysis result. Contains a single failed check identifying the name was uninspected for performance reasons. The result of other checks, including results of label analysis or canonical name are not returned.
    """

    risk_count: Literal[1] = Field(description='The number of checks that have a status of `alert` or `warn`.')
    rating: Literal[Rating.ALERT]
    highest_risk: Literal[UNINSPECTED_CHECK_RESULT] = Field(
        description='The check considered to be the highest risk. If no check has a status of `alert` or `warn`, this field is `null`.',
    )
    labels: Literal[None]
    canonical_name: Literal[None]
    inspected: Literal[False]


class BulkNameGuardBulkReport(BaseModel):
    """
    Bulk name analysis results.
    """

    results: list[ConsolidatedNameGuardReport]


class ConfusableGuardReport(ConsolidatedGraphemeGuardReport):
    is_canonical: bool = Field(description='Whether this confusable is canonical for the grapheme.')


class GraphemeGuardReport(ConsolidatedGraphemeGuardReport):
    checks: list[GenericCheckResult] = Field(description='A list of checks that were performed on the grapheme.')

    grapheme_link: Optional[str] = Field(
        description='Link to an external page with information about the grapheme.\n'
        '* `null` for multi-character graphemes',
        examples=['https://unicodeplus.com/U+0076'],
    )

    @computed_field(
        description='The name of the webpage that `grapheme_link` links to.\n'
        '* "No link is available" if `grapheme_link` is `null`'
    )
    @property
    def grapheme_link_name(self) -> str:
        if self.grapheme_link is None:
            return 'No link is available'
        else:
            return detect_grapheme_link_name(self.grapheme_link)

    @computed_field(
        description='The codepoints of the grapheme in the format `U+XXXX`. '
        'Some graphemes may have multiple codepoints.'
    )
    @property
    def codepoints(self) -> list[str]:
        return [f'U+{ord(c):04X}' for c in self.grapheme]

    confusables: list[ConfusableGuardReport] = Field(
        description='A list graphemes that can be confused with the analyzed grapheme. '
        'The list does not contain the analyzed grapheme. '
        'A canonical form of the grapheme is the first element of the list, if it is known. '
        'Otherwise, the first element is a different confusable.'
    )

    canonical_grapheme: Optional[str] = Field(
        description='A grapheme that is the canonical form of the analyzed grapheme.\n'
        '* `null` if the canonical form is not known\n'
        '* does not imply that the canonical grapheme/label/name is normalized',
        examples=['v'],
    )


class SecurePrimaryNameStatus(str, Enum):
    """
    The status of a secure primary ENS name lookup performed by NameGuard.

    * `normalized`: The ENS primary name was found and it is normalized.
    * `no_primary_name`: The ENS primary name was not found.
    * `unnormalized`: The ENS primary name was found, but it is not normalized.
    * `uninspected`: The name was exceptionally long and was not inspected for performance reasons.
    """

    NORMALIZED = 'normalized'
    NO_PRIMARY_NAME = 'no_primary_name'
    UNNORMALIZED = 'unnormalized'
    UNINSPECTED = 'uninspected'


class ImpersonationEstimate(str, Enum):
    """
    The predicted impersonation estimate of an ENS name.

    * `unlikely`: The ENS name is unlikely to be impersonating.
    * `potential`: The ENS name is potentially impersonating.
    """

    UNLIKELY = 'unlikely'
    POTENTIAL = 'potential'


class SecurePrimaryNameResult(BaseModel):
    """
    Reverse lookup result.
    """

    primary_name_status: SecurePrimaryNameStatus

    impersonation_estimate: Optional[ImpersonationEstimate] = Field(
        description='Impersonation estimate of the `primary_name`.\n' '* `null` if `primary_name` is `null`',
    )

    primary_name: Optional[str] = Field(
        description='Primary ENS name for the Ethereum address.\n'
        '* `null` if `primary_name_status` is any value except `normalized`',
        examples=['vitalik.eth'],
    )

    display_name: str = Field(
        description='ENS beautified version of `primary_name`.\n'
        '* if `primary_name` is `null` then provides a fallback `display_name` of "Unnamed [first four hex digits of Ethereum address]", e.g. "Unnamed c2a6"',
        examples=['vitalik.eth'],
    )

    nameguard_report: Optional[NameGuardReport] = Field(
        description='NameGuard report for the `primary_name`.\n'
        '* if `address` has a primary name and `return_nameguard_report` is `True`, then return a `NameGuardReport` for the primary name. Else, NameGuard will return `null`',
    )


class FakeEthNameCheckStatus(str, Enum):
    """
    * `authentic_eth_name` The NFT is associated with authentic ".eth" contracts.
    * `impersonated_eth_name` The NFT appears to impersonate a ".eth" name. It doesn't belong to authentic ENS contracts but contains graphemes that visually resemble ".eth" at the end of relevant NFT metadata fields. Consider automated rejection of this NFT from marketplaces.
    * `potentially_impersonated_eth_name` The NFT potentially impersonates a ".eth" name. It doesn't belong to authentic ENS contracts but contains graphemes that visually resemble ".eth" within relevant NFT metadata fields (but not at the end of those fields). Consider manual review of this NFT before publishing to marketplaces.
    * `non_impersonated_eth_name` The NFT doesn't represent itself as a ".eth" name and doesn't belong to authentic ENS contracts. No string that visually resembles ".eth" was found within relevant NFT metadata fields.
    * `unknown_nft`: No information could be found on the requested NFT. This generally indicates that the NFT doesn't exist or hasn't been indexed yet.
    * `invalid_eth_name`: The NFT is associated with authentic ".eth" contracts, but it is unnormalized.
    * `potentially_authentic_eth_name`: The NFT is associated with authentic ".eth" contracts, but its label is unknown.
    * `unknown_eth_name`: The NFT is associated with authentic ".eth" contracts, but its label is unknown or has never been registered.
    """

    AUTHENTIC_ETH_NAME = 'authentic_eth_name'
    IMPERSONATED_ETH_NAME = 'impersonated_eth_name'
    POTENTIALLY_IMPERSONATED_ETH_NAME = 'potentially_impersonated_eth_name'
    NON_IMPERSONATED_ETH_NAME = 'non_impersonated_eth_name'
    UNKNOWN_NFT = 'unknown_nft'
    INVALID_ETH_NAME = 'invalid_eth_name'
    UNKNOWN_ETH_NAME = 'unknown_eth_name'


class FakeEthNameCheckResult(BaseModel):
    """
    Fake .eth ENS name check result.
    """

    status: FakeEthNameCheckStatus

    nameguard_report: Optional[NameGuardReport] = Field(
        description='NameGuard report for the .eth ENS NFT.\n'
        '* `null` if `status` is any value except `authentic_eth_name`, `invalid_eth_name` and `unknown_eth_name` (the NFT is not associated with authentic ".eth" contracts)\n'
    )

    investigated_fields: Optional[dict[str, str]] = Field(
        description='Fields with values from Alchemy response which are investigated (e.g. title, collection name, metadata) whether they look like fake .eth ENS name.\n'
        '* `null` if `status` is `unknown_nft`'
    )



from typing import Optional
from nameguard.models import CheckStatus, Check, GenericCheckResult, NameCheckResult, UNINSPECTED_SKIP_MESSAGE
from label_inspector.models import InspectorResult


# title: NameWrapper Fuses
TITLE_INFO = 'NameWrapper Fuses'
TITLE_SKIP = 'NameWrapper Fuses'

MESSAGE_INFO = 'NameWrapper fuse checks will be added soon'


def check_name(labels: list[Optional[InspectorResult]]) -> GenericCheckResult:
    return NameCheckResult(
        check=Check.NAMEWRAPPER_FUSES,
        status=CheckStatus.INFO,
        _name_message=MESSAGE_INFO,
        _title=TITLE_INFO,
    )


UNINSPECTED_SKIP_CHECK_RESULT = NameCheckResult(
    check=Check.NAMEWRAPPER_FUSES,
    status=CheckStatus.SKIP,
    _name_message=UNINSPECTED_SKIP_MESSAGE,
    _title=TITLE_SKIP,
)



from .checks import (
    GenericCheckResult,
    GraphemeCheckResult,
    LabelCheckResult,
    NameCheckResult,
    Rating,
    Check,
    CheckStatus,
    UNINSPECTED_SKIP_MESSAGE,
)

from .result import (
    NameGuardReport,
    ConsolidatedNameGuardReport,
    LabelGuardReport,
    ConsolidatedGraphemeGuardReport,
    UninspectedNameGuardReport,
    BulkNameGuardBulkReport,
    ConsolidatedReport,
    Normalization,
    GraphemeNormalization,
    GraphemeGuardReport,
    SecurePrimaryNameResult,
    SecurePrimaryNameStatus,
    FakeEthNameCheckStatus,
    FakeEthNameCheckResult,
    ImpersonationEstimate,
    ConfusableGuardReport,
)

from .request import (
    NetworkName,
)



import os
import re
from typing import Union

from nameguard.models.checks import UNINSPECTED_CHECK_RESULT
from nameguard.models.result import UninspectedNameGuardReport
from nameguard.our_ens import OurENS
from ens_normalize import is_ens_normalized, ens_cure, DisallowedSequence, ens_process

import requests
from label_inspector.inspector import Inspector
from label_inspector.config import initialize_inspector_config
from label_inspector.models import InspectorConfusableGraphemeResult, InspectorResult
from web3 import HTTPProvider
from web3.exceptions import ContractLogicError
from dotenv import load_dotenv

from nameguard import checks
from nameguard.models import (
    CheckStatus,
    Check,
    NameGuardReport,
    LabelGuardReport,
    ConsolidatedGraphemeGuardReport,
    BulkNameGuardBulkReport,
    Normalization,
    GraphemeNormalization,
    GraphemeGuardReport,
    NetworkName,
    SecurePrimaryNameResult,
    SecurePrimaryNameStatus,
    FakeEthNameCheckStatus,
    FakeEthNameCheckResult,
    ImpersonationEstimate,
    ConsolidatedNameGuardReport,
    Rating,
    ConfusableGuardReport,
)
from nameguard.provider import get_nft_metadata
from nameguard.utils import (
    namehash_from_name,
    labelhash_from_label,
    calculate_nameguard_rating,
    count_risks,
    agg_checks,
    get_highest_risk,
    label_is_labelhash,
    compute_canonical_from_list,
    is_labelhash_eth,
    MAX_INSPECTED_NAME_CHARACTERS,
    labelhash_in_name,
)
from nameguard.exceptions import ProviderUnavailable, NotAGrapheme, MissingTitle
from nameguard.logging import logger
from nameguard.subgraph import (
    namehash_to_name_lookup,
    resolve_all_labelhashes_in_name_querying_labelhashes,
    resolve_all_labelhashes_in_names_querying_labelhashes,
)
from nameguard.generic_utils import capitalize_words

DNA_CHECKS = [
    (checks.dna.normalized.check_grapheme, checks.dna.normalized.check_label, checks.dna.normalized.check_name),
    (None, checks.dna.punycode.check_label, checks.dna.punycode.check_name),
]

GRAPHEME_CHECKS = [
    checks.grapheme.confusables.check_grapheme,
    checks.grapheme.font_support.check_grapheme,
    checks.grapheme.invisible.check_grapheme,
    checks.grapheme.typing_difficulty.check_grapheme,
]

LABEL_CHECKS = [
    checks.label.mixed_scripts.check_label,
    checks.label.namewrapper.check_label,
    checks.label.unknown.check_label,
]

NAME_CHECKS = [
    checks.name.impersonation_risk.check_name,
    checks.name.namewrapper_fuses.check_name,
    checks.name.decentralized_name.check_name,
]


def init_inspector():
    with initialize_inspector_config('prod_config') as config:
        if os.getenv('AWS_LAMBDA_INITIALIZATION_TYPE', None) == 'provisioned-concurrency':
            config.inspector.lazy_loading = False
        return Inspector(config)


ens_contract_adresses = {
    '0x57f1887a8bf19b14fc0df6fd9b2acc9af147ea85',  # Base Registrar
    '0xd4416b13d2b3a9abae7acd5d6c2bbdbe25686401',  # Name Wrapper
}

ALCHEMY_UNKNOWN_NAME = re.compile('^\[0x[0-9a-f]{4}\.\.\.[0-9a-f]{4}\]\.eth$')


def nested_get(dic, keys):
    for key in keys:
        dic = dic[key]
    return dic


SIMPLE_NAME_REGEX = re.compile(r'^[a-z0-9]{1,63}\.eth$')  # 63 label length limit comes from Punycode


def simple_name(name: str) -> bool:
    return SIMPLE_NAME_REGEX.match(name) is not None


def consolidated_report_from_simple_name(name: str) -> ConsolidatedNameGuardReport:
    return ConsolidatedNameGuardReport(
        name=name,
        namehash=namehash_from_name(name),
        normalization=Normalization.NORMALIZED,
        rating=Rating.PASS,
        risk_count=0,
        highest_risk=None,
        inspected=True,
    )


Uninspected_name_checks = [
    UNINSPECTED_CHECK_RESULT,
    checks.dna.normalized.UNINSPECTED_SKIP_CHECK_RESULT,
    checks.dna.punycode.UNINSPECTED_SKIP_CHECK_RESULT,
    checks.grapheme.confusables.UNINSPECTED_SKIP_CHECK_RESULT,
    checks.grapheme.font_support.UNINSPECTED_SKIP_CHECK_RESULT,
    checks.grapheme.invisible.UNINSPECTED_SKIP_CHECK_RESULT,
    checks.grapheme.typing_difficulty.UNINSPECTED_SKIP_CHECK_RESULT,
    checks.label.mixed_scripts.UNINSPECTED_SKIP_CHECK_RESULT,
    checks.label.namewrapper.UNINSPECTED_SKIP_CHECK_RESULT,
    checks.label.unknown.UNINSPECTED_SKIP_CHECK_RESULT,
    checks.name.decentralized_name.UNINSPECTED_SKIP_CHECK_RESULT,
    checks.name.impersonation_risk.UNINSPECTED_SKIP_CHECK_RESULT,
    checks.name.namewrapper_fuses.UNINSPECTED_SKIP_CHECK_RESULT,
]


def consolidated_report_from_uninspected_name(name: str) -> UninspectedNameGuardReport:
    res = ens_process(name, do_normalize=True, do_beautify=True)
    beautified = res.beautified
    normalized = name == res.normalized

    return UninspectedNameGuardReport(
        name=name,
        namehash=namehash_from_name(name),
        normalization=Normalization.UNKNOWN
        if labelhash_in_name(name)
        else Normalization.NORMALIZED
        if normalized
        else Normalization.UNNORMALIZED,
        rating=Rating.ALERT,
        risk_count=1,
        highest_risk=UNINSPECTED_CHECK_RESULT,
        beautiful_name=beautified,  # TODO computed twice
        checks=Uninspected_name_checks,
        labels=None,
        canonical_name=None,
        inspected=False,
    )


class NameGuard:
    def __init__(self):
        self._inspector = init_inspector()
        load_dotenv()
        # TODO use web sockets and async
        self.ns = {}
        for network_name, env_var in (
            (NetworkName.MAINNET, 'PROVIDER_URI_MAINNET'),
            (NetworkName.SEPOLIA, 'PROVIDER_URI_SEPOLIA'),
        ):
            if os.environ.get(env_var) is None:
                logger.warning(f'Environment variable {env_var} is not set')
            self.ns[network_name] = OurENS(HTTPProvider(os.environ.get(env_var)))

        # optimization
        self.eth_label = self._inspector.analyse_label('eth', simple_confusables=True, omit_cure=True)

    def analyse_label(self, label: str) -> InspectorResult:
        if label == 'eth':
            return self.eth_label
        return self._inspector.analyse_label(label, simple_confusables=True, omit_cure=True)

    async def inspect_name(
        self, network_name: NetworkName, name: str, resolve_labelhashes: bool = True, bulk_mode: bool = False
    ) -> Union[NameGuardReport, ConsolidatedNameGuardReport]:
        """
        Inspect a name. A name is a sequence of labels separated by dots.
        A label can be a labelhash or a string.
        If a labelhash is encountered and `resolve_labelhashes` is `True`, a lookup will be performed.
        Returns UninspectedNameGuardReport if name was exceptionally long and was not inspected for performance reasons.
        """

        logger.debug(f"[inspect_name] name: '{name}'")

        if resolve_labelhashes:
            resolved_name = await resolve_all_labelhashes_in_name_querying_labelhashes(network_name, name)
            if resolved_name is None:
                return consolidated_report_from_uninspected_name(name)
            name = resolved_name
        return self.inspect_name_sync(name, bulk_mode)

    def inspect_name_sync(
        self, name: str, bulk_mode: bool = False
    ) -> Union[NameGuardReport, ConsolidatedNameGuardReport]:
        """
        Inspect a name. A name is a sequence of labels separated by dots.
        A label can be a labelhash or a string.
        If a labelhash is encountered, it will be treated as an unknown label.
        """

        logger.debug(f"[inspect_name] name: '{name}'")

        if len(name) > MAX_INSPECTED_NAME_CHARACTERS:
            return consolidated_report_from_uninspected_name(name)

        if bulk_mode and simple_name(name):
            return consolidated_report_from_simple_name(name)

        labels = [] if len(name) == 0 else name.split('.')
        logger.debug(f'[inspect_name] labels: {labels}')

        # labelhashes have `None` as their analysis
        labels_analysis = [
            self.analyse_label(label)
            # do not analyze labelhashes
            if not label_is_labelhash(label)
            else None
            for label in labels
        ]

        # -- check individual entities --

        # checks for each grapheme in each label
        labels_graphemes_checks = [
            [[check(grapheme) for check in GRAPHEME_CHECKS] for grapheme in label_analysis.graphemes]
            if label_analysis is not None
            else []
            # label has [] graphemes if it's a labelhash
            for label_analysis in labels_analysis
        ]

        # checks for each label
        labels_checks = [
            [check(label_analysis) for check in LABEL_CHECKS]
            # checks have to handle labelhashes
            for label_analysis in labels_analysis
        ]

        # checks for the whole name
        name_checks = [check(labels_analysis) for check in NAME_CHECKS]

        # -- aggregate results --

        # merge grapheme checks into label checks
        for label_i, label_graphemes_checks in enumerate(labels_graphemes_checks):
            for grapheme_checks in label_graphemes_checks:
                labels_checks[label_i].extend([c.raise_context() for c in grapheme_checks])
            labels_checks[label_i] = agg_checks(labels_checks[label_i])

        # merge label checks into name checks
        for label_checks in labels_checks:
            name_checks.extend([c.raise_context() for c in label_checks])
        name_checks = agg_checks(name_checks)

        # -- DNA checks --

        # After the aggregation of all previous checks, run the Do-Not-Aggregate checks.
        # These checks are run on every entity (grapheme/label/name) separately.

        for check_g, check_l, check_n in DNA_CHECKS:
            for label_i, label_analysis in enumerate(labels_analysis):
                if label_analysis is not None:
                    if check_g:
                        for grapheme_i, grapheme in enumerate(label_analysis.graphemes):
                            labels_graphemes_checks[label_i][grapheme_i].append(check_g(grapheme))
                labels_checks[label_i].append(check_l(label_analysis))
            name_checks.append(check_n(labels_analysis))

        # -- generate result --

        return NameGuardReport(
            name=name,
            namehash=namehash_from_name(name),
            normalization=Normalization.UNKNOWN
            if any(label_analysis is None for label_analysis in labels_analysis)
            else Normalization.NORMALIZED
            if all(
                label_analysis.status == 'normalized' and len(label_analysis.label) > 0
                for label_analysis in labels_analysis
            )
            else Normalization.UNNORMALIZED,
            rating=calculate_nameguard_rating(name_checks),
            risk_count=count_risks(name_checks),
            highest_risk=get_highest_risk(name_checks),
            checks=sorted(name_checks, reverse=True),
            canonical_name=compute_canonical_from_list(
                [
                    label_analysis.normalized_canonical_label if label_analysis is not None else labels[i]  # labelhash
                    for i, label_analysis in enumerate(labels_analysis)
                ],
                sep='.',
            ),
            labels=[
                LabelGuardReport(
                    # actual label or [labelhash]
                    label=label,
                    labelhash=labelhash_from_label(label_analysis.label)
                    if label_analysis is not None
                    else '0x' + label[1:-1],
                    normalization=Normalization.UNKNOWN
                    if label_analysis is None
                    else Normalization.NORMALIZED
                    if label_analysis.status == 'normalized' and len(label_analysis.label) > 0
                    else Normalization.UNNORMALIZED,
                    rating=calculate_nameguard_rating(label_checks),
                    risk_count=count_risks(label_checks),
                    highest_risk=get_highest_risk(label_checks),
                    checks=sorted(label_checks, reverse=True),
                    canonical_label=label_analysis.normalized_canonical_label
                    if label_analysis is not None
                    else label,  # labelhash
                    graphemes=[
                        ConsolidatedGraphemeGuardReport(
                            normalization=GraphemeNormalization.NORMALIZED
                            if any(
                                check.status == CheckStatus.PASS and check.check is Check.NORMALIZED
                                for check in grapheme_checks
                            )
                            else GraphemeNormalization.UNNORMALIZED,
                            grapheme=grapheme.value,
                            grapheme_name=capitalize_words(grapheme.name),
                            grapheme_type=grapheme.type,
                            grapheme_script=grapheme.script,
                            grapheme_link=grapheme.link,
                            rating=calculate_nameguard_rating(grapheme_checks),
                            risk_count=count_risks(grapheme_checks),
                            highest_risk=get_highest_risk(grapheme_checks),
                            grapheme_description=grapheme.description,
                            unicode_version=grapheme.unicode_version,
                        )
                        for grapheme, grapheme_checks in zip(label_analysis.graphemes, label_graphemes_checks)
                    ]
                    if label_analysis is not None
                    else None,
                )
                for label, label_analysis, label_checks, label_graphemes_checks in zip(
                    labels,
                    labels_analysis,
                    labels_checks,
                    labels_graphemes_checks,
                )
            ],
            inspected=True,
        )

    async def bulk_inspect_names(self, network_name: NetworkName, names: list[str]) -> BulkNameGuardBulkReport:
        names = await resolve_all_labelhashes_in_names_querying_labelhashes(network_name, names)
        return BulkNameGuardBulkReport(
            results=[
                await self.inspect_name(network_name, name, resolve_labelhashes=False, bulk_mode=True) for name in names
            ],
        )

    async def inspect_namehash(self, network_name: NetworkName, namehash: str) -> NameGuardReport:
        logger.debug(f"[inspect_namehash] namehash: '{namehash}'")
        name = await namehash_to_name_lookup(network_name, namehash)
        return await self.inspect_name(network_name, name)

    async def inspect_name_with_labelhash_lookup(self, network_name: NetworkName, name: str) -> NameGuardReport:
        """
        Inspect a name. A name is a sequence of labels separated by dots.
        A label can be a labelhash or a string.
        If a labelhash is encountered, the entire name will be looked up by its namehash.
        If the namehash is not found, all labelhashes will be treated as unknown labels.
        """

        logger.debug(f"[inspect_name_with_labelhash_lookup] name: '{name}'")

        return await self.inspect_name(network_name, name)

    def inspect_grapheme(self, grapheme: str) -> GraphemeGuardReport:
        """
        Inspect a single grapheme.
        Throws `NotAGrapheme` if the input is not a single grapheme.
        """

        label_analysis = self.analyse_label(grapheme)
        if label_analysis.grapheme_length != 1:
            raise NotAGrapheme(f'The input contains {label_analysis.grapheme_length} graphemes.')

        grapheme_analysis = label_analysis.graphemes[0]
        grapheme_checks = [check(grapheme_analysis) for check in GRAPHEME_CHECKS + [c[0] for c in DNA_CHECKS if c[0]]]

        if grapheme_analysis.confusables_canonical:
            canonical = self._inspect_confusable(grapheme_analysis.confusables_canonical)
            canonical.is_canonical = True
        else:
            canonical = None

        return GraphemeGuardReport(
            normalization=GraphemeNormalization.NORMALIZED
            if any(check.status == CheckStatus.PASS and check.check is Check.NORMALIZED for check in grapheme_checks)
            else GraphemeNormalization.UNNORMALIZED,
            grapheme=grapheme_analysis.value,
            grapheme_name=capitalize_words(grapheme_analysis.name),
            grapheme_type=grapheme_analysis.type,
            grapheme_script=grapheme_analysis.script,
            grapheme_link=grapheme_analysis.link,
            rating=calculate_nameguard_rating(grapheme_checks),
            risk_count=count_risks(grapheme_checks),
            highest_risk=get_highest_risk(grapheme_checks),
            checks=sorted(grapheme_checks, reverse=True),
            confusables=([canonical] if canonical else [])
            + (
                [self._inspect_confusable(c) for c in grapheme_analysis.confusables_other]
                if grapheme_analysis.confusables_other
                else []
            ),
            canonical_grapheme=label_analysis.canonical_label,
            grapheme_description=grapheme_analysis.description,
            unicode_version=grapheme_analysis.unicode_version,
        )

    def _inspect_confusable(self, grapheme: InspectorConfusableGraphemeResult) -> ConfusableGuardReport:
        grapheme_checks = [check(grapheme) for check in GRAPHEME_CHECKS + [c[0] for c in DNA_CHECKS if c[0]]]
        return ConfusableGuardReport(
            normalization=GraphemeNormalization.NORMALIZED
            if any(check.status == CheckStatus.PASS and check.check is Check.NORMALIZED for check in grapheme_checks)
            else GraphemeNormalization.UNNORMALIZED,
            grapheme=grapheme.value,
            grapheme_name=capitalize_words(grapheme.name),
            grapheme_type=grapheme.type,
            grapheme_script=grapheme.script,
            grapheme_link=grapheme.link,
            rating=calculate_nameguard_rating(grapheme_checks),
            risk_count=count_risks(grapheme_checks),
            highest_risk=get_highest_risk(grapheme_checks),
            grapheme_description=grapheme.description,
            unicode_version=grapheme.unicode_version,
            is_canonical=False,
        )

    async def secure_primary_name(
        self, address: str, network_name: str, return_nameguard_report: bool = False
    ) -> SecurePrimaryNameResult:
        try:
            domain = self.ns[network_name].name(address)
        except requests.exceptions.ConnectionError as ex:
            raise ProviderUnavailable(f'Communication error with provider occurred: {ex}')
        except ContractLogicError:
            domain = None
        except Exception:
            domain = None
        display_name = f'Unnamed {address[2:6].lower()}'
        primary_name = None
        nameguard_report = None
        if domain is None:
            status = SecurePrimaryNameStatus.NO_PRIMARY_NAME
            impersonation_estimate = None
        else:
            nameguard_report = await self.inspect_name(network_name, domain)

            if nameguard_report.highest_risk and nameguard_report.highest_risk.check.name == Check.UNINSPECTED.name:
                status = SecurePrimaryNameStatus.UNINSPECTED
                impersonation_estimate = None
            elif nameguard_report.normalization == Normalization.UNNORMALIZED:
                status = SecurePrimaryNameStatus.UNNORMALIZED
                impersonation_estimate = None
            else:
                display_name = nameguard_report.beautiful_name
                status = SecurePrimaryNameStatus.NORMALIZED
                primary_name = domain

                impersonation_estimate = (
                    ImpersonationEstimate.UNLIKELY
                    if any(
                        check.check == 'impersonation_risk' and check.status == CheckStatus.PASS
                        for check in nameguard_report.checks
                    )
                    else ImpersonationEstimate.POTENTIAL
                )

        return SecurePrimaryNameResult(
            primary_name=primary_name,
            impersonation_estimate=impersonation_estimate,
            display_name=display_name,
            primary_name_status=status,
            nameguard_report=nameguard_report if return_nameguard_report else None,
        )

    async def fake_eth_name_check_fields(
        self, network_name, contract_address, token_id, investigated_fields: dict[str, str]
    ) -> FakeEthNameCheckResult:
        contract_address = contract_address.lower()

        return await self._fake_eth_name_check(network_name, contract_address, investigated_fields)

    async def fake_eth_name_check(self, network_name, contract_address, token_id) -> FakeEthNameCheckResult:
        """
        Check if the token is a fake ENS name (not valid ENS contract address and title and collection name of NFT look like ENS name).
        """
        contract_address = contract_address.lower()

        res_json = await get_nft_metadata(network_name, contract_address, token_id)

        token_type = res_json['id']['tokenMetadata']['tokenType']

        if token_type not in ['ERC721', 'ERC1155'] and contract_address in ens_contract_adresses:
            return FakeEthNameCheckResult(
                status=FakeEthNameCheckStatus.UNKNOWN_NFT, nameguard_report=None, investigated_fields=None
            )
        if token_type == 'NOT_A_CONTRACT':
            return FakeEthNameCheckResult(
                status=FakeEthNameCheckStatus.UNKNOWN_NFT, nameguard_report=None, investigated_fields=None
            )
        elif token_type == 'NO_SUPPORTED_NFT_STANDARD':
            return FakeEthNameCheckResult(
                status=FakeEthNameCheckStatus.UNKNOWN_NFT, nameguard_report=None, investigated_fields=None
            )
        elif token_type not in ['ERC721', 'ERC1155']:  # Alchemy does not support other types
            return FakeEthNameCheckResult(
                status=FakeEthNameCheckStatus.UNKNOWN_NFT, nameguard_report=None, investigated_fields=None
            )

        title = res_json['title']
        investigated_fields = {}
        for keys in [
            ['title'],
            ['metadata', 'name'],
            ['contractMetadata', 'openSea', 'collectionName'],
            ['contractMetadata', 'name'],
        ]:
            try:
                name = nested_get(res_json, keys)
                investigated_fields['.'.join(keys)] = name
            except KeyError:
                pass

        if contract_address in ens_contract_adresses:
            if ALCHEMY_UNKNOWN_NAME.match(title):
                unknown_name = f"[{res_json['id']['tokenId'][2:]}].eth"
                investigated_fields['title'] = unknown_name
            elif title == '':  # the name has never been registered
                investigated_fields['title'] = None

        return await self._fake_eth_name_check(network_name, contract_address, investigated_fields)

    async def _fake_eth_name_check(
        self, network_name, contract_address, fields: dict[str, str]
    ) -> FakeEthNameCheckResult:
        if contract_address in ens_contract_adresses:
            if 'title' not in fields:
                raise MissingTitle()

            title = fields['title']

            if title is None:
                return FakeEthNameCheckResult(
                    status=FakeEthNameCheckStatus.UNKNOWN_NFT, nameguard_report=None, investigated_fields=None
                )
            else:
                if is_labelhash_eth(title):
                    report = await self.inspect_name(network_name, title, resolve_labelhashes=False)
                    return FakeEthNameCheckResult(
                        status=FakeEthNameCheckStatus.UNKNOWN_ETH_NAME,
                        nameguard_report=report,
                        investigated_fields=None,
                    )

                # TODO: check if token_id matches contract_address and title
                report = await self.inspect_name(network_name, title)
                if is_ens_normalized(title):
                    return FakeEthNameCheckResult(
                        status=FakeEthNameCheckStatus.AUTHENTIC_ETH_NAME,
                        nameguard_report=report,
                        investigated_fields=None,
                    )
                else:
                    return FakeEthNameCheckResult(
                        status=FakeEthNameCheckStatus.INVALID_ETH_NAME,
                        nameguard_report=report,
                        investigated_fields=None,
                    )
        else:
            impersonating_fields = {}
            impersonated = False
            potentially_impersonated = False
            for key, name in fields.items():
                try:
                    cured_title = ens_cure(name)  # TODO improve, e.g. remove invisible and then canonicalize
                    inspector_result = self.analyse_label(cured_title)
                    if inspector_result.canonical_label is not None:
                        canonical_name = inspector_result.canonical_label
                    else:
                        canonical_name = cured_title
                except DisallowedSequence:
                    canonical_name = name

                if canonical_name.endswith('.eth'):
                    impersonating_fields[key] = name
                    impersonated = True
                elif '.eth' in canonical_name:
                    impersonating_fields[key] = name
                    potentially_impersonated = True

            if impersonated:
                return FakeEthNameCheckResult(
                    status=FakeEthNameCheckStatus.IMPERSONATED_ETH_NAME,
                    nameguard_report=None,
                    investigated_fields=impersonating_fields,
                )
            elif potentially_impersonated:
                return FakeEthNameCheckResult(
                    status=FakeEthNameCheckStatus.POTENTIALLY_IMPERSONATED_ETH_NAME,
                    nameguard_report=None,
                    investigated_fields=impersonating_fields,
                )
            else:
                return FakeEthNameCheckResult(
                    status=FakeEthNameCheckStatus.NON_IMPERSONATED_ETH_NAME,
                    nameguard_report=None,
                    investigated_fields=None,
                )



import gc
import json
import os
import time

# ruff: noqa: E402
init_time = time.time()

from enum import Enum
from fastapi import FastAPI, Path
from fastapi.middleware.cors import CORSMiddleware
from pydantic import BaseModel, Field

import nameguard.context
from nameguard.nameguard import NameGuard
from nameguard.utils import (
    validate_namehash,
    namehash_from_labelhash,
    validate_token_id,
    validate_ethereum_address,
    MAX_NUMBER_OF_NAMES_IN_BULK,
)
from nameguard.models import (
    NameGuardReport,
    BulkNameGuardBulkReport,
    SecurePrimaryNameResult,
    NetworkName,
    GraphemeGuardReport,
    FakeEthNameCheckResult,
)
from nameguard.endpoints import Endpoints
from nameguard.logging import logger
from nameguard.exceptions import (
    InvalidNameHash,
    ENSSubgraphUnavailable,
    NamehashMismatchError,
    NamehashNotFoundInSubgraph,
    InvalidEthereumAddress,
    ProviderUnavailable,
    InvalidTokenID,
    NotAGrapheme,
    MissingTitle,
)

logger.debug('NameGuard starting.')


class ApiVersion(str, Enum):
    V08_BETA = 'v0.8-beta'


app = FastAPI(
    title='NameGuard API',
    version=ApiVersion.V08_BETA.value,
    description="""Welcome to NameGuard, a powerful tool developed by [NameHash Labs](https://namehashlabs.org) to identify and prevent malicious use of Ethereum Name Service (ENS) names.

## Important Notes
⚠️ Beta Version: Please be aware that NameGuard is currently in beta. We appreciate your feedback as we work towards a stable v1 release. Expect changes to the API to address community feedback and improve functionality.

## Documentation
These documentation pages focus specifically on the NameGuard API. For information on the NameGuard Library, SDK, and UI Kit, please refer to the [NameKit GitHub repository](https://github.com/namehash/namekit).

## Glossary

* uninspected name - any name that fails any of the following requirements:
  1. The count of unknown labels (distinct or non-distinct) must not exceed `MAX_INSPECTED_NAME_UNKNOWN_LABELS`.
  2. The count of characters in the name (after potential resolving of unknown labels) must not exceed the limit `MAX_INSPECTED_NAME_CHARACTERS`.
If name is uninspected a `NameGuardReport` with `inspected` field equal to `false` is returned. 

""",
    servers=[
        {'url': 'https://api.nameguard.io', 'description': 'Production server'},
        {'url': '/', 'description': 'Default, relative server'},
    ],
)

app.add_middleware(
    CORSMiddleware,
    allow_origins=['*'],
    allow_credentials=True,
    allow_methods=['*'],
    allow_headers=['*'],
)

init_call = True


class LogEntry:
    def __init__(self):
        self.start_time = time.time()
        self.lambda_type = os.getenv('AWS_LAMBDA_INITIALIZATION_TYPE', None)

    def log(self, data: dict):
        global init_call
        data['processing_time'] = time.time() - self.start_time
        data['init_call'] = init_call
        init_call = False
        data['time_since_init'] = time.time() - init_time
        data['lambda_type'] = self.lambda_type

        logger.debug(f'{json.dumps(data)}')


ng = NameGuard()
logger.debug('NameGuard inited.')

if os.getenv('AWS_LAMBDA_INITIALIZATION_TYPE', None) == 'provisioned-concurrency':
    ng.analyse_label('ni ck.eth')
    logger.debug('NameGuard warmup.')


# -- inspect-name --


class InspectNameRequest(BaseModel):
    name: str = Field(
        description='Name to inspect.',
        examples=['vitalìk.eth'],
    )
    network_name: NetworkName


@app.get(
    '/inspect-name/{network_name}/{name:path}',
    tags=['name'],
    summary='Inspect Name',
    responses={
        **ENSSubgraphUnavailable.get_responses_spec(),
    },
)
async def inspect_name_get(
    network_name: NetworkName,
    name: str = Path(
        description='**Name should be url-encoded (except when using the Swagger UI). Name can be empty.**',
        examples=['vitalìk.eth'],
    ),
) -> NameGuardReport:
    """
    ## Inspects a single name with NameGuard.

    If the `name` is uninspected returns `NameGuardReport` with `inspected` field equal to `false`; else returns a `NameGuardReport` including:
    1. The details of all checks performed on `name` that consolidates all checks performed on labels and graphemes in `name`.
    2. The details of all labels in `name`.
    3. A consolidated inspection result of all graphemes in `name`.

    This endpoint will attempt automated labelhash resolution through the ENS Subgraph,
    using the network specified in `network_name`.
    """

    log_entry = LogEntry()
    nameguard.context.endpoint_name.set(Endpoints.INSPECT_NAME)
    result = await ng.inspect_name(network_name, name)

    log_entry.log(
        {
            'endpoint': Endpoints.INSPECT_NAME,
            'method': 'GET',
            'network_name': network_name,
            'name': name,
        }
    )

    return result


@app.get(
    '/inspect-name/{network_name}',
    include_in_schema=False,
    tags=['name'],
    summary='Inspect Empty Name',
    responses={
        **ENSSubgraphUnavailable.get_responses_spec(),
    },
)
async def inspect_empty_name_get(network_name: NetworkName) -> NameGuardReport:
    return await inspect_name_get(network_name, '')


@app.post(
    '/inspect-name',
    tags=['name'],
    summary='Inspect Name',
    responses={
        **ENSSubgraphUnavailable.get_responses_spec(),
    },
)
async def inspect_name_post(request: InspectNameRequest) -> NameGuardReport:
    """
    ## Inspects a single name with NameGuard.

    If the `name` is uninspected returns `NameGuardReport` with `inspected` field equal to `false`; else returns a `NameGuardReport` including:
    1. The details of all checks performed on `request.name` that consolidates all checks performed on labels and graphemes in `request.name`.
    2. The details of all labels in `request.name`.
    3. A consolidated inspection result of all graphemes in `request.name`.

    This endpoint will attempt automated labelhash resolution through the ENS Subgraph,
    using the network specified in `request.network_name`.
    """

    log_entry = LogEntry()
    nameguard.context.endpoint_name.set(Endpoints.INSPECT_NAME)
    result = await ng.inspect_name(request.network_name, request.name)

    log_entry.log(
        {
            'endpoint': Endpoints.INSPECT_NAME,
            'method': 'POST',
            'network_name': request.network_name,
            'name': request.name,
        }
    )

    return result


# -- bulk-inspect-names --


class BulkInspectNamesRequest(BaseModel):
    names: list[str] = Field(max_length=MAX_NUMBER_OF_NAMES_IN_BULK, examples=[['vitalìk.eth', 'nick.eth']])
    network_name: NetworkName


@app.post(
    '/bulk-inspect-names',
    tags=['name'],
    summary='Inspect Multiple Names',
    responses={
        **ENSSubgraphUnavailable.get_responses_spec(),
    },
)
async def bulk_inspect_names(request: BulkInspectNamesRequest) -> BulkNameGuardBulkReport:
    """
    ## Inspects up to 250 names at a time with NameGuard.

    Provides `NameGuardReport` with `inspected` field equal to `false` if the `name` is uninspected; else returns a `ConsolidatedNameGuardReport` for each name provided in `request.names`, including:
    1. The details of all checks performed on a name that consolidates all checks performed on labels and graphemes in this name.

    Each `ConsolidatedNameGuardReport` returned represents an equivalent set of checks as a `NameGuardReport`. However:
    1. A `NameGuardReport` contains a lot of additional data that isn't always needed / desired when a `ConsolidatedNameGuardReport` will do.
    2. When NameGuard only needs to return a `ConsolidatedNameGuardReport`, some special performance optimizations are possible (and completely safe) that help to accelerate responses in many cases.

    This endpoint will attempt automated labelhash resolution through the ENS Subgraph,
    using the network specified in `request.network_name`.
    """

    log_entry = LogEntry()
    nameguard.context.endpoint_name.set(Endpoints.BULK_INSPECT_NAMES)
    result = await ng.bulk_inspect_names(request.network_name, request.names)

    log_entry.log(
        {
            'endpoint': Endpoints.BULK_INSPECT_NAMES,
            'method': 'POST',
            'network_name': request.network_name,
            'names': request.names,
            'names_count': len(request.names),
        }
    )

    return result


# -- inspect-namehash --


class InspectNamehashRequest(BaseModel):
    namehash: str = Field(
        description='Namehash should be a decimal or a hex (prefixed with 0x) string.',
        examples=['0xd48fd5598e605861cbd8e45419b41b83739bff52eaef0e283181bbe0a43a5b32'],
    )
    network_name: NetworkName


@app.get(
    '/inspect-namehash/{network_name}/{namehash}',
    tags=['namehash'],
    summary='Inspect Namehash GET',
    responses={
        **InvalidNameHash.get_responses_spec(),
        **ENSSubgraphUnavailable.get_responses_spec(),
        **NamehashMismatchError.get_responses_spec(),
        **NamehashNotFoundInSubgraph.get_responses_spec(),
    },
)
async def inspect_namehash_get(
    network_name: NetworkName,
    namehash: str = Path(
        examples=['0xd48fd5598e605861cbd8e45419b41b83739bff52eaef0e283181bbe0a43a5b32'],
        description='Namehash should be a decimal or a hex (prefixed with 0x) string.',
    ),
) -> NameGuardReport:
    """
    ## Inspects the name associated with a namehash.

    NameGuard will attempt to resolve the name associated with the `namehash` through the ENS Subgraph,
    using the network specified in `network_name`.

    If this resolution succeeds then NameGuard will return a `NameGuardReport` for the name. If the resolved `name` is uninspected then NameGuard will return `NameGuardReport` with `inspected` field equal to `false`.

    If this resolution fails then NameGuard will return an error.
    """

    logger.debug(
        f"{json.dumps({'endpoint': Endpoints.INSPECT_NAMEHASH, 'method': 'GET', 'network_name': network_name, 'namehash': namehash})}"
    )
    nameguard.context.endpoint_name.set(Endpoints.INSPECT_NAMEHASH)
    return await ng.inspect_namehash(
        network_name=network_name,
        namehash=validate_namehash(namehash=namehash),
    )


@app.post(
    '/inspect-namehash',
    tags=['namehash'],
    summary='Inspect Namehash',
    responses={
        **InvalidNameHash.get_responses_spec(),
        **ENSSubgraphUnavailable.get_responses_spec(),
        **NamehashMismatchError.get_responses_spec(),
        **NamehashNotFoundInSubgraph.get_responses_spec(),
    },
)
async def inspect_namehash_post(request: InspectNamehashRequest) -> NameGuardReport:
    """
    ## Inspects the name associated with a namehash.

    NameGuard will attempt to resolve the name associated with the `request.namehash` through the ENS Subgraph,
    using the network specified in `request.network_name`.

    If this resolution succeeds then NameGuard will return a `NameGuardReport` for the name. If the resolved `name` is uninspected then NameGuard will return `NameGuardReport` with `inspected` field equal to `false`.

    If this resolution fails then NameGuard will return an error.
    """

    logger.debug(
        f"{json.dumps({'endpoint': Endpoints.INSPECT_NAMEHASH, 'method': 'POST', 'network_name': request.network_name, 'namehash': request.namehash})}"
    )
    nameguard.context.endpoint_name.set(Endpoints.INSPECT_NAMEHASH)
    return await ng.inspect_namehash(
        network_name=request.network_name,
        namehash=validate_namehash(namehash=request.namehash),
    )


# -- inspect-labelhash --


class InspectLabelhashRequest(BaseModel):
    labelhash: str = Field(
        description='Labelhash should be a decimal or a hex (prefixed with 0x) string.',
        examples=['0x3276e4878615389906712b876ce1455b8f5d1c5ea3ffcf7a705e0d32fafae9c5'],
    )
    network_name: NetworkName
    parent_name: str = Field('eth')


@app.get(
    '/inspect-labelhash/{network_name}/{labelhash}/{parent_name:path}',
    tags=['labelhash'],
    summary='Inspect Labelhash GET',
    responses={
        **InvalidNameHash.get_responses_spec(),
        **ENSSubgraphUnavailable.get_responses_spec(),
        **NamehashMismatchError.get_responses_spec(),
        **NamehashNotFoundInSubgraph.get_responses_spec(),
    },
)
async def inspect_labelhash_get(
    network_name: NetworkName,
    labelhash: str = Path(
        examples=['0x3276e4878615389906712b876ce1455b8f5d1c5ea3ffcf7a705e0d32fafae9c5'],
        description='Labelhash should be a decimal or a hex (prefixed with 0x) string.',
    ),
    parent_name: str = Path(
        description='Parent name of the labelhash (default empty).',
        examples=['eth'],
    ),
) -> NameGuardReport:
    """
    ## Inspects the name `[{labelhash}].{parent_name}`.

    Parent may be a name with any number of labels. The default parent is "eth".

    This is a convenience endpoint to generate a `NameGuardReport` in cases when you only have:
    1. The labelhash of the "childmost" label of a name.
    2. The complete parent name of the "childmost" label.

    Returns `NameGuardReport` with `inspected` field equal to `false` if the resolved name is uninspected.

    NameGuard always inspects names, rather than labelhashes. So this endpoint will first attempt
    to resolve the "childmost" label associated with the provided labelhash through the ENS Subgraph,
    using the network specified in `network_name`.

    If this label resolution fails the resulting `NameGuardReport` will be equivalent to requesting
    a `NameGuardReport` for the name `[{labelhash}].{parent_name}` which will contain (at least)
    one label with an `unknown` `Normalization`.

    If this label resolution succeeds the resulting `NameGuardReport` will be equivalent to requesting
    a `NameGuardReport` for the name `{label}.{parent_name}`.
    """

    logger.debug(
        f"{json.dumps({'endpoint': Endpoints.INSPECT_LABELHASH, 'method': 'GET', 'network_name': network_name, 'labelhash': labelhash, 'parent_name': parent_name})}"
    )
    nameguard.context.endpoint_name.set(Endpoints.INSPECT_LABELHASH)
    valid_labelhash = validate_namehash(namehash=labelhash)
    namehash = namehash_from_labelhash(valid_labelhash, parent_name=parent_name)
    return await ng.inspect_namehash(network_name=network_name, namehash=namehash)


@app.get(
    '/inspect-labelhash/{network_name}/{labelhash}',
    include_in_schema=False,
    tags=['labelhash'],
    summary='Inspect Labelhash GET',
    responses={
        **InvalidNameHash.get_responses_spec(),
        **ENSSubgraphUnavailable.get_responses_spec(),
        **NamehashMismatchError.get_responses_spec(),
        **NamehashNotFoundInSubgraph.get_responses_spec(),
    },
)
async def inspect_labelhash_get_empty_parent(
    network_name: NetworkName,
    labelhash: str = Path(
        examples=['0x3276e4878615389906712b876ce1455b8f5d1c5ea3ffcf7a705e0d32fafae9c5'],
        description='Labelhash should be a decimal or a hex (prefixed with 0x) string.',
    ),
) -> NameGuardReport:
    return await inspect_labelhash_get(network_name, labelhash, '')


@app.post(
    '/inspect-labelhash',
    tags=['labelhash'],
    summary='Inspect Labelhash',
    responses={
        **InvalidNameHash.get_responses_spec(),
        **ENSSubgraphUnavailable.get_responses_spec(),
        **NamehashMismatchError.get_responses_spec(),
        **NamehashNotFoundInSubgraph.get_responses_spec(),
    },
)
async def inspect_labelhash_post(request: InspectLabelhashRequest) -> NameGuardReport:
    """
    ## Inspects the name `[{request.labelhash}].{request.parent_name}`.

    Parent may be a name with any number of labels. The default parent is "eth".

    This is a convenience endpoint to generate a `NameGuardReport` in cases when you only have:
    1. The labelhash of the "childmost" label of a name.
    2. The complete parent name of the "childmost" label.

    Returns `NameGuardReport` with `inspected` field equal to `false` if the resolved name is uninspected.

    NameGuard always inspects names, rather than labelhashes. So this endpoint will first attempt
    to resolve the "childmost" label associated with the provided labelhash through the ENS Subgraph,
    using the network specified in `request.network_name`.

    If this label resolution fails the resulting `NameGuardReport` will be equivalent to requesting
    a `NameGuardReport` for the name `[{request.labelhash}].{request.parent_name}` which will contain (at least)
    one label with an `unknown` `Normalization`.

    If this label resolution succeeds the resulting `NameGuardReport` will be equivalent to requesting
    a `NameGuardReport` for the name `{label}.{request.parent_name}`.
    """

    logger.debug(
        f"{json.dumps({'endpoint': Endpoints.INSPECT_LABELHASH, 'method': 'POST', 'network_name': request.network_name, 'labelhash': request.labelhash, 'parent_name': request.parent_name})}"
    )
    nameguard.context.endpoint_name.set(Endpoints.INSPECT_LABELHASH)
    valid_labelhash = validate_namehash(namehash=request.labelhash)
    namehash = namehash_from_labelhash(valid_labelhash, parent_name=request.parent_name)
    return await ng.inspect_namehash(network_name=request.network_name, namehash=namehash)


# -- secure-primary-name --


@app.get(
    '/secure-primary-name/{network_name}/{address}',
    tags=['secure_primary_name'],
    summary='Reverse lookup of an Ethereum address to a primary name',
    responses={
        **InvalidEthereumAddress.get_responses_spec(),
        **ProviderUnavailable.get_responses_spec(),
    },
)
async def secure_primary_name_get(
    address: str, network_name: NetworkName, return_nameguard_report: bool = False
) -> SecurePrimaryNameResult:
    """
    ## Performs a reverse lookup of an Ethereum `address` to a primary name.

    Data sources for the primary name lookup include:
    1. The Ethereum Provider configured in the NameGuard instance.
    2. For ENS names using CCIP-Read: requests to externally defined gateway servers.

    Returns `display_name` to be shown to users and estimates `impersonation_estimate`.

    If `address` has a primary name and `return_nameguard_report` is `True`, then NameGuard will return a `SecurePrimaryNameResult` including a `NameGuardReport` for the primary name. Else, NameGuard will return `None` as `nameguard_report`.
    """
    logger.debug(
        f"{json.dumps({'endpoint': Endpoints.SECURE_PRIMARY_NAME, 'method': 'GET', 'network_name': network_name, 'address': address, 'return_nameguard_report': return_nameguard_report})}"
    )
    nameguard.context.endpoint_name.set(Endpoints.SECURE_PRIMARY_NAME)
    address = validate_ethereum_address(address)

    return await ng.secure_primary_name(address, network_name, return_nameguard_report)


# -- fake-ens-name-check --


class FakeETHNameCheckFieldsRequest(BaseModel):
    network_name: NetworkName
    contract_address: str = Path(
        examples=['0x495f947276749ce646f68ac8c248420045cb7b5e'],
        description='Contract address for the NFT contract (ERC721 and ERC1155 supported).',
    )
    token_id: str = Path(
        examples=['61995921128521442959106650131462633744885269624153038309795231243542768648193'],
        description='The ID of the token (in hex or decimal format).',
    )
    fields: dict[str, str] = Field(
        description='Fields with values which will be investigated (e.g. title, collection name, metadata) whether they look like fake .eth ENS name. `title` key is mandatory, for ENS contracts it should be the ENS name.'
    )


@app.post(
    '/fake-eth-name-check',
    tags=['fake-eth-name-check'],
    summary='Fake .eth ENS name check with fields',
    responses={
        **InvalidTokenID.get_responses_spec(),
        **InvalidEthereumAddress.get_responses_spec(),
        **MissingTitle.get_responses_spec(),
    },
)
async def fake_eth_name_check_fields_post(request: FakeETHNameCheckFieldsRequest) -> FakeEthNameCheckResult:
    """
    ## Performs a fake .eth ENS name check based on given NFT metadata.

    This endpoint checks if the metadata of an NFT looks like a fake .eth ENS name.
    """
    logger.debug(
        f"{json.dumps({'endpoint': Endpoints.FAKE_ETH_NAME_CHECK, 'method': 'POST', 'network_name': request.network_name, 'contract_address': request.contract_address, 'token_id': request.token_id})}"
    )
    nameguard.context.endpoint_name.set(Endpoints.FAKE_ETH_NAME_CHECK)
    contract_address = validate_ethereum_address(request.contract_address)
    token_id = validate_token_id(request.token_id)
    return await ng.fake_eth_name_check_fields(
        network_name=request.network_name,
        contract_address=contract_address,
        token_id=token_id,
        investigated_fields=request.fields,
    )


# -- inspect-grapheme --


@app.get(
    '/inspect-grapheme/{grapheme}',
    tags=['grapheme'],
    summary='Inspect Grapheme GET',
    responses={
        **NotAGrapheme.get_responses_spec(),
    },
)
async def inspect_grapheme_get(
    grapheme: str = Path(
        description='Grapheme to inspect. Should be url-encoded (except when using the Swagger UI).',
        examples=['v', 'ń', '%F0%9F%98%B5'],
    ),
) -> GraphemeGuardReport:
    """## Inspects a single grapheme."""
    logger.debug(f"{json.dumps({'endpoint': Endpoints.INSPECT_GRAPHEME, 'method': 'GET', 'grapheme': grapheme})}")
    nameguard.context.endpoint_name.set(Endpoints.INSPECT_GRAPHEME)
    return ng.inspect_grapheme(grapheme)


if __name__ == '__main__':
    import asyncio

    asyncio.run(ng.inspect_name('mainnet', 'ni ck.eth'))


gc.freeze()



from label_inspector.models import InspectorGraphemeWithConfusablesResult as Grapheme
from nameguard.models import CheckStatus, Check, GenericCheckResult, GraphemeCheckResult, UNINSPECTED_SKIP_MESSAGE

STATUS = CheckStatus.WARN

# title: Font Support
TITLE_PASS = 'Font Support'
TITLE_FAIL = 'Font Support'
TITLE_SKIP = 'Font Support'

MESSAGE_PASS = 'Commonly supported'

MESSAGE_FAIL = 'Less common support'

MESSAGE_SKIP = 'Unknown font support'


def check_grapheme(grapheme: Grapheme) -> GenericCheckResult:
    passed = grapheme.font_support_all_os
    if passed is None:
        return GraphemeCheckResult(
            check=Check.FONT_SUPPORT,
            status=CheckStatus.SKIP,
            _grapheme_message=MESSAGE_SKIP,
            _label_message=MESSAGE_SKIP,
            _name_message=MESSAGE_SKIP,
            _title=TITLE_SKIP,
        )
    else:
        return GraphemeCheckResult(
            check=Check.FONT_SUPPORT,
            status=CheckStatus.PASS if passed else STATUS,
            _grapheme_message=MESSAGE_PASS if passed else MESSAGE_FAIL,
            _label_message=MESSAGE_PASS if passed else MESSAGE_FAIL,
            _name_message=MESSAGE_PASS if passed else MESSAGE_FAIL,
            _title=TITLE_PASS if passed else TITLE_FAIL,
        )


UNINSPECTED_SKIP_CHECK_RESULT = GraphemeCheckResult(
    check=Check.FONT_SUPPORT,
    status=CheckStatus.SKIP,
    _grapheme_message=UNINSPECTED_SKIP_MESSAGE,
    _label_message=UNINSPECTED_SKIP_MESSAGE,
    _name_message=UNINSPECTED_SKIP_MESSAGE,
    _title=TITLE_SKIP,
)



from typing import Optional
from label_inspector.models import InspectorResult
from nameguard.models import CheckStatus, Check, GenericCheckResult, LabelCheckResult, UNINSPECTED_SKIP_MESSAGE

STATUS = CheckStatus.WARN

# title: NameWrapper Compatible
TITLE_PASS = 'NameWrapper Compatibility'
TITLE_FAIL = 'NameWrapper Compatibility'
TITLE_SKIP = 'NameWrapper Compatibility'

L_MESSAGE_PASS = 'Compatible for use with the ENS NameWrapper'
N_MESSAGE_PASS = 'Compatible for use with the ENS NameWrapper'

L_MESSAGE_FAIL = 'Incompatible with the ENS NameWrapper'
N_MESSAGE_FAIL = 'Incompatible with the ENS NameWrapper'

L_MESSAGE_SKIP = 'This label is unknown'
N_MESSAGE_SKIP = 'This name contains unknown labels'

WRAPPED_MAX_BYTES = 255


def check_label(label: Optional[InspectorResult]) -> GenericCheckResult:
    if label is None:
        return LabelCheckResult(
            check=Check.NAMEWRAPPER_COMPATIBLE,
            status=CheckStatus.SKIP,
            _label_message=L_MESSAGE_SKIP,
            _name_message=N_MESSAGE_SKIP,
            _title=TITLE_SKIP,
        )
    try:
        wrapped = label.label.encode('utf-8')
        passed = len(wrapped) <= WRAPPED_MAX_BYTES
    except UnicodeEncodeError:
        passed = False
    return LabelCheckResult(
        check=Check.NAMEWRAPPER_COMPATIBLE,
        status=CheckStatus.PASS if passed else STATUS,
        _label_message=L_MESSAGE_PASS if passed else L_MESSAGE_FAIL,
        _name_message=N_MESSAGE_PASS if passed else N_MESSAGE_FAIL,
        _title=TITLE_PASS if passed else TITLE_FAIL,
    )


UNINSPECTED_SKIP_CHECK_RESULT = LabelCheckResult(
    check=Check.NAMEWRAPPER_COMPATIBLE,
    status=CheckStatus.SKIP,
    _label_message=UNINSPECTED_SKIP_MESSAGE,
    _name_message=UNINSPECTED_SKIP_MESSAGE,
    _title=TITLE_SKIP,
)



# import hashlib
# import json
from typing import Optional
import os
from dotenv import load_dotenv


import httpx

from nameguard.logging import logger
from nameguard.exceptions import ENSSubgraphUnavailable, NamehashNotFoundInSubgraph, NamehashMismatchError
from nameguard.models import NetworkName
from nameguard.utils import namehash_from_name, label_is_labelhash, MAX_INSPECTED_NAME_UNKNOWN_LABELS

load_dotenv()

# The label limit for using the multi-label lookup query.
# Longer names will be resolved by querying the namehash of the full name.
MAX_MULTI_LABEL_LOOKUP = 256

ENS_SUBGRAPH_URL = {
    NetworkName.MAINNET: os.environ.get('ENS_SUBGRAPH_URL_MAINNET'),
    NetworkName.SEPOLIA: os.environ.get('ENS_SUBGRAPH_URL_SEPOLIA'),
}


RESOLVE_NAMEHASH_QUERY = """
query resolveNamehash($nameHash: String) {
  domain(id: $nameHash) {
    name
  }
}
"""

headers = {'Content-Type': 'application/json'}


async def call_subgraph(network_name: NetworkName, query: str, variables: dict) -> dict:
    try:
        async with httpx.AsyncClient() as client:
            response = await client.post(
                ENS_SUBGRAPH_URL[network_name] + '?source=ens-nameguard',
                json={'query': query, 'variables': variables},
                headers=headers,
            )

        if response.status_code == 200:
            response_json = response.json()
            logger.debug(f'Subgraph query:\n{query} {variables}')
            logger.debug(f'Subgraph response json:\n{response_json}')
        else:
            raise ENSSubgraphUnavailable(
                f'Received unexpected status code from ENS Subgraph {response.status_code}: {response.text}'
            )
    except Exception as ex:
        logger.exception('Communication error with subgraph occurred')
        if not str(ex):
            raise ENSSubgraphUnavailable('Communication error with subgraph occurred.')
        raise ENSSubgraphUnavailable(f'Communication error with subgraph occurred: {ex}')

    if 'data' not in response_json:
        logger.error(f'Unexpected response body: {response_json}')
        raise ENSSubgraphUnavailable(f'Unexpected response body: {response_json}')
    else:
        return response_json['data']


async def namehash_to_name_lookup(network_name: NetworkName, namehash_hexstr: str) -> str:
    logger.debug(f'Trying namehash lookup for: {namehash_hexstr}')
    namehash_hexstr = namehash_hexstr.lower()

    variables = {'nameHash': namehash_hexstr}

    data = await call_subgraph(network_name, RESOLVE_NAMEHASH_QUERY, variables)

    if 'domain' not in data:
        logger.error(f'Unexpected response data: {data}')
        raise ENSSubgraphUnavailable(f'Unexpected response data: {data}')
    elif data['domain'] is None:
        raise NamehashNotFoundInSubgraph()
    elif 'name' in data['domain']:
        if data['domain']['name'] is None:
            logger.error(f'Name by namehash is None: {data} {variables}')
            name = ''
        else:
            name = str(data['domain']['name'])
        calculated_namehash = namehash_from_name(name)
        if calculated_namehash != namehash_hexstr:
            logger.error(
                f'NamehashMismatchError occurred:\ninput: {namehash_hexstr}\tcalculated: {calculated_namehash}'
            )
            raise NamehashMismatchError()
        return name
    else:
        logger.error(f'Unexpected response data: {data}')
        raise ENSSubgraphUnavailable(f'Unexpected response data: {data}')


def build_multi_label_query(labels: list[str]) -> tuple[str, dict]:
    """
    Builds a query that resolves all labelhashes in a name.

    ```
    query resolveLabelhashes($l0: String, $l1: String) {
        l0: domain(id: $l0) {
            labelName
        }
        l1: domain(id: $l1) {
            labelName
        }
    }
    ```
    """
    labelhash_idx = [i for i, label in enumerate(labels) if label_is_labelhash(label)]
    args = ' '.join(f'$l{i}:String' for i in labelhash_idx)
    query = f'query resolveLabelhashes({args}){{'
    variables = {}
    for i in labelhash_idx:
        partial_name = '.'.join(labels[i:])
        partial_namehash = namehash_from_name(partial_name)
        query += f'l{i}:domain(id:$l{i}){{labelName}}'
        variables[f'l{i}'] = partial_namehash
    query += '}'
    return query, variables


async def resolve_all_labelhashes_in_name(network_name: NetworkName, name: str) -> str:
    logger.debug(f'Trying to resolve full name: {name}')

    namehash = namehash_from_name(name)
    labels = name.split('.')

    if len(labels) < MAX_MULTI_LABEL_LOOKUP:
        logger.debug(f'Trying multi-label lookup for: {name}')
        query, variables = build_multi_label_query(labels)
        data = await call_subgraph(network_name, query, variables)
        resolved_labels = []
        for i, label in enumerate(labels):
            if label_is_labelhash(label):
                # domain could be unknown or just the label could be unknown
                if data[f'l{i}'] is not None:
                    label = data[f'l{i}']['labelName'] or label
            resolved_labels.append(label)
        resolved_name = '.'.join(resolved_labels)
        resolved_namehash = namehash_from_name(resolved_name)
        if resolved_namehash != namehash:
            logger.error(f'NamehashMismatchError occurred:\ninput: {name}\tcalculated: {resolved_name}')
            raise NamehashMismatchError()
    else:
        logger.debug(f'Trying namehash lookup for: {name}')
        try:
            resolved_name = await namehash_to_name_lookup(network_name, namehash)
        except NamehashNotFoundInSubgraph:
            resolved_name = name

    logger.debug(f'Resolved name: {resolved_name}')

    return resolved_name


def build_multi_label_query_querying_labelhashes(labelhashes: list[str]) -> tuple[str, dict]:
    """
    Builds a query that resolves all labelhashes in a name.
    Labelhash in format 0x1234...1234.
    ```
    query resolveLabelhashes($l0: String, $l1: String) {
        l0: domains(where: {labelhash: $l0, labelName_not: null}, first: 1) {
            labelName
        }
        l1: domains(where: {labelhash: $l1, labelName_not: null}, first: 1) {
            labelName
        }
    }
    ```
    """
    args = ' '.join(f'$l{i}:String' for i in range(len(labelhashes)))
    query = f'query resolveLabelhashes({args}){{'
    variables = {}
    for i, labelhash in enumerate(labelhashes):
        query += f'l{i}:domains(where: {{labelhash: $l{i}, labelName_not: null}}, first: 1){{labelName}}'
        variables[f'l{i}'] = labelhash
    query += '}'
    return query, variables


async def resolve_labelhashes_querying_labelhashes(network_name: NetworkName, labelhashes: list[str]) -> dict[str, str]:
    """
    Resolve labelhashes to label names.
    Labelhash in format [1234...1234].
    """
    if not labelhashes:
        return {}
    labelhashes = [f'0x{labelhash[1:-1]}' for labelhash in labelhashes]
    query, variables = build_multi_label_query_querying_labelhashes(labelhashes)

    data = await call_subgraph(network_name, query, variables)

    result = {}
    for var, labelhash in variables.items():
        labelhash = f'[{labelhash[2:]}]'
        if data[var]:
            result[labelhash] = data[var][0]['labelName']
        else:
            result[labelhash] = labelhash
    return result


async def resolve_all_labelhashes_in_name_querying_labelhashes(network_name: NetworkName, name: str) -> Optional[str]:
    """
    Resolve up to MAX_INSPECTED_NAME_UNKNOWN_LABELS labelhashes in a name. If there is more labelhashes then return None.
    """
    labels = name.split('.')

    labelhash_idx = [i for i, label in enumerate(labels) if label_is_labelhash(label)]

    if not labelhash_idx:
        return name

    if len(labelhash_idx) > MAX_INSPECTED_NAME_UNKNOWN_LABELS:
        return None

    logger.debug(f'Trying to resolve full name: {name}')
    resolved_labelhashes = await resolve_labelhashes_querying_labelhashes(
        network_name, [labels[i] for i in labelhash_idx]
    )

    for i in labelhash_idx:
        labels[i] = resolved_labelhashes[labels[i]]

    resolved_name = '.'.join(labels)

    logger.debug(f'Resolved name: {resolved_name}')

    return resolved_name


async def resolve_all_labelhashes_in_names_querying_labelhashes(
    network_name: NetworkName, names: list[str]
) -> list[str]:
    segmented_names = []
    labelhashes = set()
    labelhashes_list = []
    for name in names:
        labels = name.split('.')
        segmented_names.append(labels)
        for label in labels:
            if label_is_labelhash(label):
                if label not in labelhashes:
                    labelhashes.add(label)
                    labelhashes_list.append(label)

    resolved_labelhashes = await resolve_labelhashes_querying_labelhashes(network_name, labelhashes_list)

    resolved_names = []
    for labels in segmented_names:
        for i, label in enumerate(labels):
            if label_is_labelhash(label):
                labels[i] = resolved_labelhashes[label]
        resolved_names.append('.'.join(labels))
    return resolved_names



# The MIT License (MIT)
#
# Copyright (c) 2016 The Ethereum Foundation
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# import json
from functools import wraps
from typing import Optional, Union, Tuple, cast

from ens.constants import ENS_EXTENDED_RESOLVER_INTERFACE_ID, EMPTY_SHA3_BYTES
from ens.ens import _resolver_supports_interface, ENS
from ens.exceptions import ENSValidationError
from ens.utils import address_to_reverse_domain, is_empty_name, is_none_or_zero_address, Web3

# from ens.utils import normalize_name
from eth_typing import ChecksumAddress
from eth_utils import to_checksum_address, is_address, to_bytes
from hexbytes import HexBytes
from web3.contract import Contract


def label_to_hash(label: str) -> HexBytes:
    # label = normalize_name(label)
    if '.' in label:
        raise ValueError(f"Cannot generate hash for label {label!r} with a '.'")
    return Web3().keccak(text=label)


def normal_name_to_hash(name: str) -> HexBytes:
    """
    This method will not normalize the name. 'normal' name here means the name
    should already be normalized before calling this method.

    :param name:            the name to hash - should already be normalized
    :return: namehash       the hash of the name
    :rtype: HexBytes
    """
    node = EMPTY_SHA3_BYTES
    if not is_empty_name(name):
        labels = name.split('.')
        for label in reversed(labels):
            labelhash = label_to_hash(label)
            assert isinstance(labelhash, bytes)
            assert isinstance(node, bytes)
            node = Web3().keccak(node + labelhash)
    return node


def raw_name_to_hash(name: str) -> HexBytes:
    """
    Generate the namehash. This is also known as the ``node`` in ENS contracts.

    In normal operation, generating the namehash is handled
    behind the scenes. For advanced usage, it is a helpful utility.

    This normalizes the name with `nameprep
    <https://github.com/ethereum/EIPs/blob/master/EIPS/eip-137.md#name-syntax>`_
    before hashing.

    :param str name: ENS name to hash
    :return: the namehash
    :rtype: bytes
    :raises InvalidName: if ``name`` has invalid syntax
    """
    # normalized_name = normalize_name(name)
    normalized_name = name
    return normal_name_to_hash(normalized_name)


def ens_encode_name(name: str) -> bytes:
    """
    Encode a name according to DNS standards specified in section 3.1
    of RFC1035 with the following validations:

        - There is no limit on the total length of the encoded name
        and the limit on labels is the ENS standard of 255.

        - Return a single 0-octet, b'\x00', if empty name.

    :param str name: the dot-separated ENS name
    """
    if is_empty_name(name):
        return b'\x00'

    # normalized_name = normalize_name(name)
    normalized_name = name

    labels = normalized_name.split('.')
    labels_as_bytes = [to_bytes(text=label) for label in labels]

    # raises if len(label) > 255:
    for index, label in enumerate(labels):
        if len(label) > 255:
            raise ENSValidationError(f'Label at position {index} too long after encoding.')

    # concat label size in bytes to each label:
    dns_prepped_labels = [to_bytes(len(label)) + label for label in labels_as_bytes]

    # return the joined prepped labels in order and append the zero byte at the end:
    return b''.join(dns_prepped_labels) + b'\x00'


class OurENS(ENS):
    """
    Name normalization is disabled.
    """

    def name(self, address: ChecksumAddress) -> Optional[str]:
        """
        Look up the name that the address points to, using a
        reverse lookup. Reverse lookup is opt-in for name owners.

        :param address:
        :type address: hex-string
        """
        reversed_domain = address_to_reverse_domain(address)
        name = self._resolve(reversed_domain, fn_name='name')

        # To be absolutely certain of the name, via reverse resolution,
        # the address must match in the forward resolution
        result = name if to_checksum_address(address) == self.address(name) else None

        return result

    def resolver(self, name: str) -> Optional['Contract']:
        """
        Get the resolver for an ENS name.

        :param str name: The ENS name
        """
        # normal_name = normalize_name(name)
        normal_name = name
        return self._get_resolver(normal_name)[0]

    @staticmethod
    @wraps(raw_name_to_hash)
    def namehash(name: str) -> HexBytes:
        return raw_name_to_hash(name)

    def _resolve(self, name: str, fn_name: str = 'addr') -> Optional[Union[ChecksumAddress, str]]:
        # normal_name = normalize_name(name)
        normal_name = name
        resolver, current_name = self._get_resolver(normal_name, fn_name)
        if not resolver:
            return None

        node = self.namehash(normal_name)

        # handle extended resolver case
        if _resolver_supports_interface(resolver, ENS_EXTENDED_RESOLVER_INTERFACE_ID):
            contract_func_with_args = (fn_name, [node])

            calldata = resolver.encodeABI(*contract_func_with_args)
            contract_call_result = resolver.caller.resolve(
                ens_encode_name(normal_name),  # TODO ens_encode_name
                calldata,
            )
            result = self._decode_ensip10_resolve_data(contract_call_result, resolver, fn_name)
            return to_checksum_address(result) if is_address(result) else result
        elif normal_name == current_name:
            lookup_function = getattr(resolver.functions, fn_name)
            result = lookup_function(node).call()
            if is_none_or_zero_address(result):
                return None
            return to_checksum_address(result) if is_address(result) else result
        return None

    def _get_resolver(
        self,
        normal_name: str,
        fn_name: str = 'addr',
    ) -> Tuple[Optional['Contract'], str]:
        current_name = normal_name

        # look for a resolver, starting at the full name and taking the parent
        # each time that no resolver is found
        while True:
            if is_empty_name(current_name):
                # if no resolver found across all iterations, current_name
                # will eventually be the empty string '' which returns here
                return None, current_name

            resolver_addr = self.ens.caller.resolver(normal_name_to_hash(current_name))
            if not is_none_or_zero_address(resolver_addr):
                # if resolver found, return it
                resolver = cast('Contract', self._type_aware_resolver(resolver_addr, fn_name))
                return resolver, current_name

            # set current_name to parent and try again
            current_name = self.parent(current_name)



import os
import httpx

from nameguard.logging import logger
from nameguard.exceptions import ProviderUnavailable
from nameguard.models import NetworkName

alchemy_uris = {
    NetworkName.MAINNET: os.environ.get('ALCHEMY_URI_MAINNET'),
    NetworkName.SEPOLIA: os.environ.get('ALCHEMY_URI_SEPOLIA'),
}


async def get_nft_metadata(network_name, contract_address: str, token_id: str) -> dict:
    url = f'{alchemy_uris[network_name]}/getNFTMetadata?contractAddress={contract_address}&tokenId={token_id}&refreshCache=false'
    headers = {'accept': 'application/json'}

    try:
        async with httpx.AsyncClient() as client:
            response = await client.get(
                url,
                headers=headers,
            )

        if response.status_code == 200:
            response_json = response.json()
            logger.debug(f'Provider response json:\n{response_json}')
            if 'error' in response_json and 'timed out' in response_json['error']:
                raise ProviderUnavailable(
                    f"Received retryable error from provider {response.status_code}: {response_json['error']}"
                )
        else:
            raise ProviderUnavailable(
                f'Received unexpected status code from provider {response.status_code}: {response.text}'
            )
    except Exception as ex:
        logger.exception('Communication error with provider occurred')
        if not str(ex):
            raise ProviderUnavailable('Communication error with provider occurred.')
        raise ProviderUnavailable(f'Communication error with provider occurred: {ex}')

    return response_json
